# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

from typing import *


# noinspection PyProtectedMember
class View(dict):
    """
    A widget build configuration for user interface definition. Used e.g. as parameters in some :py:class:`Application`
    methods.
    See also the sample applications.
    """

    # noinspection PyPep8Naming
    def __init__(self, clsname: Optional[str] = None, enabled: Optional[bool] = None, visibility: Optional[str] = None,
                 hstretch: Optional[float] = None, strictHorizontalSizing: Optional[bool] = None,
                 vstretch: Optional[float] = None, strictVerticalSizing: Optional[bool] = None,
                 styleClass: Optional[str] = None, width: Optional[str] = None, height: Optional[str] = None, **b: Any):
        """
        :param clsname: The clove widget class name. See clove documentation.
        :param enabled: If the widget is enabled (instead of being locked for user interaction).
        :param visibility: The visibility, as one of :py:class:`guibase.Visibility`.
        :param hstretch: The horizontal stretch affinity factor.
        :param strictHorizontalSizing: If to not overfulfill horizontal sizing requirements even for alignment purposes.
        :param vstretch: The vertical stretch affinity factor.
        :param strictVerticalSizing: If to not overfulfill vertical sizing requirements even for alignment purposes.
        :param styleClass: Additional CSS classes. See also :py:meth:`Application.add_clientstyle`.
        :param width: A fixed width (specified as in css).
        :param height: A fixed height (specified as in css).
        :param b: Additional configuration. Depends on the widget class.
        """
        b.update({k: v for k, v in dict(view=clsname, enabled=enabled, visibility=visibility, hstretch=hstretch,
                                        strictHorizontalSizing=strictHorizontalSizing, vstretch=vstretch,
                                        strictVerticalSizing=strictVerticalSizing, styleClass=styleClass, width=width,
                                        height=height).items() if v is not None})
        if b.get('view', '') in ["clove.PopupMenuButton", "PopupMenuButton"]:
            b['view'] = "YajiPopupMenuButton"
        super().__init__(**b)

    class _DataBinding:
        """
        A data binding.
        """

        def __init__(self, datasource: Any, data_direction: Optional[str], convertername: Optional[str]):
            """
            :param datasource: The internal data source definition.
            :param data_direction: The data direction. See :py:class:`BindDirection`.
            :param convertername: Name of the converter class.
            """
            self.__datasource = datasource
            self.__data_direction = data_direction
            self.__convertername = convertername

        def _to_simple_repr_(self):
            return '__DataBinding', {'datasource': self.__datasource, 'data_direction': self.__data_direction,
                                     'convertername': self.__convertername}

    class _EventBinding:
        """
        An event binding.
        """

        def __init__(self, function: 'AbstractFunction', curtain: Union[bool, Optional[str]]):
            """
            :param function: The function to bind to an event.
            :param curtain: The curtain expression.
            """
            self.__function = function
            self.__curtain = curtain

        def _to_simple_repr_(self):
            return '__EventBinding', {'function': self.__function, 'curtain': self.__curtain}


class _AbstractContainer(View):
    """
    Abstract container view.
    """

    def __init__(self, clsname: str, paramname: str, singlechild: bool, default: Any, *children: View, **b):
        """
        :param clsname: The clove classname.
        :param paramname: The clove property name.
        :param singlechild: If this widget is the container for only a single child widget.
        :param default: The default value.
        :param children: Child views.
        """
        b[paramname] = b.pop(paramname, None) or (children[0] if singlechild else children) or default
        super().__init__(clsname, **b)


class HorizontalStack(_AbstractContainer):
    """
    A horizontal stack view.
    """

    def __init__(self, *children: View, **b):
        """
        :param children: The child views.
        """
        super().__init__('HorizontalStack', 'cols', False, [], *children, **b)


class VerticalStack(_AbstractContainer):
    """
    A vertical stack view.
    """

    def __init__(self, *children: View, **b):
        """
        :param children: The child views.
        """
        super().__init__('VerticalStack', 'rows', False, [], *children, **b)


class Grid(_AbstractContainer):
    """
    A grid view.
    """

    def __init__(self, *children: View, **b):
        """
        :param children: The child views.
        """
        super().__init__('Grid', 'children', False, [], *children, **b)


class Wrap(_AbstractContainer):
    """
    A wrap view.
    """

    def __init__(self, *children: View, **b):
        """
        :param children: The child views.
        """
        super().__init__('Wrap', 'children', False, [], *children, **b)


class ScrollView(_AbstractContainer):
    """
    A scroll view.
    """

    def __init__(self, body: Optional[View], **b):
        """
        :param body: The child view.
        """
        super().__init__('ScrollView', 'body', True, None, body, **b)


class Spacer(View):
    """
    A spacer.
    """

    def __init__(self, **b):
        super().__init__('Spacer', **b)


class Dialog:
    """
    A dialog.
    """

    def __init__(self, ctl: 'DialogController', cfg: View, **showcfg: Any):
        """
        |infrastructure|

        :param ctl: The dialog controller.
        :param cfg: The dialog view configuration.
        :param showcfg: Additional configuration for dialog presentation.
        """
        self.__ctl = ctl
        self.__cfg = cfg
        self.__showcfg = showcfg
        self.__dialogid = None
        self.__closed_handlers = []

    @property
    def dialogid(self) -> int:
        """
        The dialog id.
        |infrastructure|
        """
        return self.__dialogid

    def show(self) -> None:
        """
        Shows the dialog.
        """
        # noinspection PyProtectedMember
        self.__ctl.show_dialog(self)

    def close(self) -> None:
        """
        Closes the dialog.
        """
        # noinspection PyProtectedMember
        self.__ctl.close_dialog(self)

    def add_closed_handler(self, fct: Callable[[], None]) -> None:
        """
        Adds a closed handler. It gets executed when the dialog closes.
        """
        self.__closed_handlers.append(fct)

    def _set_dialogid(self, dialogid: Optional[int]) -> None:
        """
        Sets the dialog id.
        """
        self.__dialogid = dialogid

    def _exec_closed_handlers(self) -> None:
        """
        Executes the closed handlers.
        """
        for fct in self.__closed_handlers:
            fct()

    def _to_simple_repr_(self):
        return 'Dialog', {'dialogid': self.dialogid, 'cfg': self.__cfg, 'showcfg': self.__showcfg}


if TYPE_CHECKING:  # ugly hack for type annotations
    from .dialogcontroller import *
    from .guibase import *


# TODO finish readme, finetune api doc, linking works?
