# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature that contains basic versatile tools. They are not associated to any particular kind of functionality.
"""

from typing import *
import importlib
import json
import os
import subprocess

import anise.framework.engine
import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


class Internals:

    @staticmethod
    def extractfilesfromretval(r: Any, autoopen: bool) -> Any:
        """
        Can be used after executing a task in the anise engine for making a result file available in the target
        directory. Before you do so, all files created as result are in a temporary area, which will be flushed when
        the process terminates.

        :param r: The raw result (arbitrary object, list, whatever).
        :param autoopen: If anise shall automatically open a file in the result if there is a hint set this way.
        """
        if isinstance(r, dict):
            result = dict()
            for k in r:
                result[k] = Internals.extractfilesfromretval(r[k], autoopen)
        elif isinstance(r, list) or isinstance(r, tuple):
            _result = list()
            for e in r:
                _result += [Internals.extractfilesfromretval(e, autoopen)]
            if isinstance(r, tuple):
                result = tuple(_result)
            else:
                result = _result
        else:
            if hasattr(r, "dl"):
                result = r.dl()
                if autoopen:
                    autoopenpath = r.getdata("autoopen", None)
                    autoopenpathterminal = r.getdata("autoopenterminal", None)
                    if autoopenpath:
                        anise.utils.basic.openurl(result + autoopenpath)
                    if autoopenpathterminal:
                        r, o = anise.utils.basic.call("less", "--version")
                        if r:
                            print("TASK RESULT:\n\n" + anise.utils.basic.readfromfile(result + autoopenpathterminal))
                        else:
                            subprocess.call(["less", result + autoopenpathterminal])
            else:
                result = r
        return result

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject() -> None:
        # set some default
        universe.name = "aniseproject"
        universe.base.authors = []
        universe.base.email = None
        universe.base.longdescription = None
        universe.base.summary = None
        universe.base.maturity = None
        universe.base.version = None
        universe.base.keywords = ""
        universe.base.basecolor = (1, 2, 4)
        universe.base.websiteurl = None
        universe.base.license = None
        universe.addtask(Internals.getprojectstatusdescription, "base.", invisible=True)
        universe.addtask(Internals.getbackgroundimage, "base.", invisible=True)
        universe.addtask(Internals.getlogoimage, "base.", invisible=True)
        universe.addtask(Internals.getregisteredtasks, "base.", invisible=True)

        def _resulthandler(result: Any) -> None:
            targetdir = getattr(universe, "targetdir", None) or "."
            if not targetdir.startswith("/"):
                targetdir = universe.specialpaths.startdir + "/" + targetdir
            os.chdir(targetdir)
            result = Internals.extractfilesfromretval(result, getattr(universe.base, "autoopenresult", True))
            if result is not None:
                anise.utils.logging.loginfo("TASK RESULT:\n\n" + str(result), printcaller=False)
        universe.addhook(anise.framework.engine.HOOK_TASK_RESULT_AVAILABLE, _resulthandler)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _beautify_longdescription() -> None:

        def hascommonprefixlength(l, lst):
            if len(lst) == 0:
                return False
            pre = lst[0][0:l]
            if len(pre) < l:
                return False
            if pre.strip() != "":
                return False
            return 0 == len([x for x in lst[1:] if not x.startswith(pre)])

        if getattr(universe.base, "longdescription", ""):
            _ls = universe.base.longdescription.split("\n")
            ls = _ls[1:-1]
            commonprefixlength = 0
            while hascommonprefixlength(commonprefixlength + 1, ls):
                commonprefixlength += 1
            for i in range(1, len(_ls) - 1):
                _ls[i] = _ls[i][commonprefixlength:]
            if _ls[-1][0:commonprefixlength].strip() == "":
                _ls[-1] = _ls[-1][commonprefixlength:]
            universe.base.longdescription = "\n".join([l.rstrip() for l in _ls])

    @staticmethod
    def getbackgroundimage() -> Optional[str]:
        """
        Determines the location of the background image.

        :return: File path relative to the project root directory.
        """
        for _bi in ["_meta/background.jpg", "_meta/background.png", "_meta/homepage_bannerimage.png"]:
            if os.path.isfile(f"{universe.specialpaths.projectdir}/{_bi}"):
                return _bi

    @staticmethod
    def getlogoimage() -> Optional[str]:
        """
        Determines the location of the background image.

        :return: File path relative to the project root directory.
        """
        for _bi in [f"_meta/{universe.name}.128.png", f"_meta/{universe.name}.png"]:
            if os.path.isfile(f"{universe.specialpaths.projectdir}/{_bi}"):
                return _bi

    @staticmethod
    def getprojectstatusdescription() -> str:
        """
        Returns a short plaintext describing the specified project maturity (:samp:`universe.base.maturity`).
        """
        if getattr(universe.base, "maturity", None):
            if universe.base.maturity == Maturity.Planning:
                w = "planning"
            elif universe.base.maturity == Maturity.PreAlpha:
                w = "pre-alpha"
            elif universe.base.maturity == Maturity.Alpha:
                w = "alpha"
            elif universe.base.maturity == Maturity.Beta:
                w = "beta"
            elif universe.base.maturity == Maturity.ProductionStable:
                w = "production-stable"
            elif universe.base.maturity == Maturity.Mature:
                w = "mature"
            elif universe.base.maturity == Maturity.Inactive:
                w = "inactive"
            else:
                raise anise.framework.exceptions.BadInputError("unknown maturity: " + universe.base.maturity)
            return f"{universe.name} is in {w} state."
        else:
            return ""

    @staticmethod
    def getregisteredtasks() -> anise.framework.files.TextFileByContent:
        """
        Determines the list of all available tasks.
        """
        res = []
        for t in universe.registeredtasks():
            res.append({'parentname': t.parentname, 'taskname': t.taskname, 'name': t.name, 'label': t.label,
                        'important': t.important, 'invisible': t.invisible, 'sortindex': t.sortindex})
        return anise.framework.files.TextFileByContent(json.dumps(res))


class Maturity:
    """
    Enumeration describing possible project states.
    """
    Planning = 'Development Status :: 1 - Planning'
    PreAlpha = 'Development Status :: 2 - Pre-Alpha'
    Alpha = 'Development Status :: 3 - Alpha'
    Beta = 'Development Status :: 4 - Beta'
    ProductionStable = 'Development Status :: 5 - Production/Stable'
    Mature = 'Development Status :: 6 - Mature'
    Inactive = 'Development Status :: 7 - Inactive'


def set_basecolor(r: float, g: float, b: float) -> None:
    """
    Sets the project base color. This color is used for all kinds of styling.

    :param r: Red value (:samp:`0.0` to :samp:`1.0`).
    :param g: Green value (:samp:`0.0` to :samp:`1.0`).
    :param b: Blue value (:samp:`0.0` to :samp:`1.0`).
    """
    universe.base.basecolor = (r, g, b)


def pyscript(code: str, hook: str, provides: Optional[List[str]] = None, requires: Optional[List[str]] = None,
             prepares: Optional[List[str]] = None) -> None:
    """
    Adds a Python function given as source code text to a hook (see Anise documentation for details about hooks).

    .. note::
       This call only has an effect, if this hook will be triggered afterwards. So, usage in a project
       description file has no effect for hooks that are triggered in very early stages!

    :param code: The Python code block to register as a hook handler.
    :param hook: The hook name (i.e. the event). This is typically a variable that begins with :samp:`HOOK_`.
    :param provides: Used for dependency ordering of hooks. A list of string symbols that this method provides.
    :param requires: Used for dependency ordering of hooks. A list of string symbols that this method requires.
                     Each method that provides one of those symbols is guaranteed to be executed before.
    :param prepares: Used for dependency ordering of hooks. A list of string symbols that this method prepares.
                     Each method that provides one of those symbols is guaranteed to be executed afterwards.
    """
    provides = provides or []
    requires = requires or []
    prepares = prepares or []
    d = dict(locals(), **globals())
    d.update(importlib.import_module("anise.framework.imports").__dict__)
    origd = dict(d)
    try:
        exec(code, d, d)
        newobjs = [x for x in list(d.keys() - origd.keys()) if not x.startswith("_")]
        if len(newobjs) != 1:
            raise anise.framework.exceptions.BadInputError(
                "pyscript/code must define exactly one object (not beginning with '_'), but defined: " + str(newobjs))
        universe.addhook(hook, d[newobjs[0]], provides=provides, requires=requires, prepares=prepares)
    except BaseException as e:
        anise.utils.logging.logerror("There were errors while loading a script.", printcaller=False)
        anise.utils.logging.logexception(e)


def pytask(code: str, taskname: str, decorators: Optional[List[str]] = None, label: Optional[str] = None) -> None:
    """
    Adds a task implementation to the universe object, so it is available for execution, e.g. directly from command
    line or from the graphical task chooser.

    :param code: The Python code block to register as task implementation.
    :param taskname: The name for this new task.
    :param decorators: Additional Python function decorators to apply.
    :param label: An optional label text.
    """
    decorators = decorators or []
    fullcode = "\n".join(decorators) + \
               f"\ndef {taskname}(*args, **kwargs):\n" + ("\n".join([" "+x for x in code.split("\n")]))
    def _exec():
        try:
            d = dict(locals(), **globals())
            d.update(importlib.import_module("anise.framework.imports").__dict__)
            exec(fullcode, d, d)
            taskfct = d[taskname]
            taskfct._aniseinternals_altname = taskname
            universe.addtask(taskfct, taskname, label=label)
        except BaseException as e:
            anise.utils.logging.logerror("There were errors while loading task '{}'.".format(taskname),
                                         printcaller=False)
            anise.utils.logging.logexception(e)
    universe.addhook(anise.framework.engine.HOOK_AFTER_PREPARATION, _exec)
