# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for compiling binaries with cmake.
"""

from typing import *
import os
import shutil

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

make = anise.framework.features.loadfeature("build.make").featuremodule


class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION, requires=["milieus"])
    def _initproject():
        universe.milieus.register("cmake", CMakeMilieuComposite)


class CMakeMilieuComposite(make.MakeMilieuComposite):
    """
    Build environment composite for Qt and qmake builds.
    
    See :py:class:`milieus.MilieuComposite` for general infos.
    """

    def __init__(self, cmakepath: str = "cmake", makepath: str = "make"):
        """
        :param cmakepath: The path to the cmake tool.
        :param makepath: The path to the make tool.
        """
        super().__init__(makepath=makepath)
        self.cmakepath = cmakepath

    def call_cmake(self, *args, buildtype='Release', **kwargs):
        """
        Calls the cmake tool.
        """
        return self.call(self.cmakepath, f"-DCMAKE_BUILD_TYPE={buildtype}",
                         raise_on_errors="unable to run cmake", *args, **kwargs)


def build(source: anise.framework.files.Filestructure, targetname: str, project: str = "/",
          milieu: Optional['milieus.Milieu'] = None,
          targetfile: Optional[str] = None) -> anise.framework.files.Filestructure:
    """
    Builds with cmake.

    :param source: The source file structure to build.
    :param targetname: The name of the make target for creating.
    :param project: The cmake project root directory path relative to :samp:`source`s root.
    :param milieu: The milieu to use.
    :param targetfile: The file name of the target file to create.
    """
    targetfile = targetfile or targetname
    with universe.executionscope.newsubscope(f"Building '{targetname}' via cmake"):
        milieu = universe.milieus.get(milieu, defaultmilieu="cmake")
        milieu.call_cmake("--version")
        with milieu.createtempdir() as tempdir:
            with anise.utils.basic.ChDir(tempdir.path):
                sourcepath = milieu.createtempdir().path
                source.dl(sourcepath)
                milieu.call_cmake(f"{universe.specialpaths.projectdir}/{project}")
                milieu.make(targetname)
                ftargetfile = f"{tempdir.path}/{project}/{targetfile}"
                xtmp = milieu.createtempdir()
                result = anise.framework.files.Filestructure(f"{xtmp.path}/{os.path.basename(targetfile)}")
                anise.framework.files.Filestructure(ftargetfile).dl(result.path)
        return result


if TYPE_CHECKING:
    import anise.features.milieus as milieus
