# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for generating all kinds of documentation like output.
"""

from typing import *
import base64
import codecs
import colorsys
import datetime
import glob
import json
import os
import subprocess
import tempfile

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.framework.report
import anise.utils

from anise.framework.projects import universe


# noinspection PyPep8Naming
class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.docgen = docgen
        universe.docrender.snippets = Internals.ShortSnippets()
        universe.docrender.mediadirs = Internals.MediaDirs()

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _initproject2():
        mf = f"{universe.specialpaths.projectdir}/_meta/media"
        if os.path.isdir(mf):
            universe.docrender.mediadirs.add(mf)

    @staticmethod
    def scalecolor(c: Tuple[float, float, float], brightness: float,
                   saturation: Optional[float] = None) -> Tuple[float, float, float]:
        """
        Internally used for scaling a color tuple to a new one with the same hue but a given brightness.
        """
        h, l, s = colorsys.rgb_to_hls(*c)
        if saturation is not None:
            s *= saturation
            if s > 1:
                s = 1.0
        return colorsys.hls_to_rgb(h, brightness, s)

    class ShortSnippets:
        """
        Storage of text snippets for use in the documentation texts like :samp:`|a_foo|`.
        """

        def __init__(self):
            self.__storage = {}
            self.add("dooverride",
                     "*Override this method in custom subclasses.*")
            self.add("dooverride_optional",
                     "*Override this method in custom subclasses or leave the default implementation.*")
            self.add("infrastructure",
                     "*This is part of a particular piece of internal infrastructure and is typically not required "
                     "to be used directly.*")

        def add(self, name: str, text: str) -> None:
            """
            Adds a text snippet to the storage, so it can be used.

            :param name: The snippet alias name. This is the :samp:`foo` in :samp:`|a_foo|`.
            :param text: The snippet substitution text.
            """
            self.__storage[name] = text

        def getall(self) -> Dict[str, str]:
            """
            Returns all aliases and their snippets.
            """
            return dict(self.__storage)

    class MediaDirs:
        """
        Storage of media directories that are used by documentation sources.
        """

        def __init__(self):
            self.__storage = []

        def add(self, path: str) -> None:
            """
            Adds a media directory.

            :param path: Path to the media directory, absolute or relative to the project directory.
            """
            self.__storage.append(path if os.path.isabs(path) else f"{universe.specialpaths.projectdir}/{path}")

        def get_mediadirs(self) -> List[str]:
            """
            Returns a list of all media directories.
            """
            return list(self.__storage)


class AbstractApiReferenceLanguage:
    """
    Base class for a programming language in api references. See :py:class:`ApiReferencePiece`.
    """

    def sphinx_config(self, rootpath: str, outpath: str) -> str:
        """
        Returns additional Sphinx :file:`conf.py` configuration.

        :param rootpath: Path to the original source directory.
        :param outpath: Path to the root directory of the Sphinx input.
        """
        return ""

    def sphinx_transfer(self, srcpath: str, outpath: str, heading: str) -> None:
        """
        Transfers auxiliary files and prepares the intermediate directory :samp:`outpath`.

        :param srcpath: The root path of the sources directory.
        :param outpath: The root path of the intermediate directory given to Sphinx.
        :param heading: The api reference heading.
        """
        anise.framework.files.Filestructure(srcpath).dl(outpath)


class Python3ApiReferenceLanguage(AbstractApiReferenceLanguage):
    """
    Python 3 language support for api references.
    """

    def patch_property_types_in_docstrings(self, pd: str) -> None:  # TODO# remove this when Sphinx can do that better
        """
        .. deprecated:: all
        """
        # noinspection PyShadowingNames
        def patchfile(p):
            # noinspection PyShadowingNames
            def patchprop(cnt, j):
                # noinspection PyShadowingNames
                def moddocstr(propname, cnt, inxl, iind, rettype, sblk):
                    linsi = cnt.find(sblk, cnt.find(sblk, inxl)+1)
                    llinsi = cnt.rfind("\n", 0, linsi) + 1 + iind
                    issettable = f"@{propname}.setter" in cnt
                    alb = "\n" + (" "*iind)
                    return cnt[:llinsi] + alb + f":rtype: {rettype}" + \
                        (f"\n{alb}This property is also settable." if issettable else "") + alb + cnt[llinsi:]
                # noinspection PyShadowingNames
                def insdocstr(propname, cnt, inxl, iind, rettype):
                    return moddocstr(propname, cnt[:inxl+iind] + 2 * ("'''\n" + (" "*iind)) + cnt[inxl+iind:],
                                     inxl, iind, rettype, "'''")
                bi = cnt.find("(", j)
                if bi != -1:
                    defi = cnt.find("def", j, bi)
                    if defi != -1:
                        propname = cnt[defi+3:bi].strip()
                        cc = 1
                        while (cc > 0) and bi < len(cnt):
                            bi += 1
                            if cnt[bi] == "(":
                                cc += 1
                            elif cnt[bi] == ")":
                                cc -= 1
                        icl = cnt.find(":", bi)
                        if icl != -1:
                            rrrettype = cnt[bi+1:icl].strip()
                            if rrrettype.startswith("->"):
                                rettype = rrrettype[2:].strip()
                                if rettype[0] in ['"', "'"]:
                                    rettype = rettype[1:-1]
                                inxl = cnt.find("\n", icl) + 1
                                if inxl != 0:
                                    iind = len(cnt[inxl:]) - len(cnt[inxl:].lstrip())
                                    sblk = cnt[inxl:].strip()[:3]
                                    for tblk in ["'''", '"""']:
                                        if sblk == tblk:
                                            cnt = moddocstr(propname, cnt, inxl, iind, rettype, sblk)
                                            break
                                    else:
                                        cnt = insdocstr(propname, cnt, inxl, iind, rettype)
                return cnt
            with open(p, 'r') as f:
                cnt = f.read()
            j = None
            while True:
                j = cnt.rfind("@property", 0, j)
                if j == -1:
                    break
                cnt = patchprop(cnt, j)
            with open(p, 'w') as f:
                f.write(cnt)
        for p in glob.glob(f"{pd}/**/*.py", recursive=True):
            patchfile(p)

    def __init__(self, show_undoc_members: bool = True, show_protected_members: bool = True,
                 exclude: Optional[List[str]] = None, preparecode: str = ""):
        self.__show_undoc_members = show_undoc_members
        self.__show_protected_members = show_protected_members
        self.__exclude = exclude or []
        self.__preparecode = preparecode

    def sphinx_config(self, rootpath, outpath):
        return (f"extensions.append('sphinx.ext.autodoc')\n"
                f"sys.path.insert(0, {repr(rootpath)})\n"
                f"{self.__preparecode}")

    def sphinx_transfer(self, srcpath, outpath, heading):
        self.patch_property_types_in_docstrings(srcpath)
        env = dict(os.environb)
        apidocopts = ["members", "show-inheritance"]
        if self.__show_undoc_members:
            apidocopts += ['undoc-members']
        env['SPHINX_APIDOC_OPTIONS'] = ','.join(apidocopts).encode()
        cmd = ['sphinx-apidoc', '--no-toc']
        if self.__show_protected_members:
            cmd += ['--private']
        cmd += ['-o', outpath, srcpath]
        for excl in self.__exclude:
            cmd += glob.glob(f"{srcpath}/{excl}", recursive=True)
        subprocess.check_call(cmd, env=env)
        with open(f'{outpath}.rst', 'w') as f:
            f.write(f"""{docgen.heading(heading)}
.. toctree::
   :glob:

   {os.path.basename(outpath)}/*
""")


class AbstractDoxygenSupportedApiReferenceLanguage(AbstractApiReferenceLanguage):
    """
    Language support for api references powered by Doxygen.
    """

    def __init__(self, *, _doxygenopts: Optional[Dict[str, str]] = None, extract_all: bool = True,
                 extract_private: bool = True, extract_package: bool = True, extract_static: bool = True,
                 extract_local_classes: bool = True, extract_local_methods: bool = True,
                 extract_anon_nspaces: bool = True, extract_priv_virtual: bool = True, file_patterns: str = "*",
                 inline_inherited_memb: bool = True, inherit_docs: bool = True,
                 hide_undoc_members: bool = False, hide_undoc_classes: bool = False,
                 exclude_patterns: str = "", predefined: Optional[List[str]] = None):
        def _b(b: bool) -> str:
            return "YES" if b else "NO"
        self.__doxygenopts = {
            'GENERATE_XML': 'YES', 'XML_OUTPUT': '_doxygen_xml', 'EXTRACT_ALL': _b(extract_all),
            'EXTRACT_PRIVATE': _b(extract_private), 'EXTRACT_PACKAGE': _b(extract_package),
            'EXTRACT_STATIC': _b(extract_static), 'EXTRACT_LOCAL_CLASSES': _b(extract_local_classes),
            'EXTRACT_LOCAL_METHODS': _b(extract_local_methods), 'EXTRACT_ANON_NSPACES': _b(extract_anon_nspaces),
            'EXTRACT_PRIV_VIRTUAL': _b(extract_priv_virtual), 'HIDE_UNDOC_MEMBERS': _b(hide_undoc_members),
            'HIDE_UNDOC_CLASSES': _b(hide_undoc_classes), 'GENERATE_HTML': 'NO', 'GENERATE_LATEX': 'NO',
            'INLINE_INHERITED_MEMB': _b(inline_inherited_memb), 'INHERIT_DOCS': _b(inherit_docs),
            'RECURSIVE': 'YES', 'FILE_PATTERNS': file_patterns, 'EXCLUDE_PATTERNS': exclude_patterns,
            'PREDEFINED': ' '.join(predefined or []), **(_doxygenopts or {})}

    def __info(self, outpath):
        project = os.path.basename(outpath)
        parpath = os.path.abspath(f"{outpath}/..")
        tmpoutpath = f"{parpath}/_tmp/{os.path.basename(outpath)}"
        doxygenxmlpath = f"{tmpoutpath}/_doxygen_xml"
        return project, tmpoutpath, doxygenxmlpath

    def sphinx_config(self, rootpath, outpath):
        anise.utils.basic.verify_tool_installed('doxygen')
        anise.utils.basic.verify_tool_installed('breathe-apidoc')
        project, tmpoutpath, doxygenxmlpath = self.__info(outpath)
        return (f"extensions.append('breathe')\n"
                f"breathe_default_members = ('members', 'private-members', 'undoc-members')\n"
                f"breathe_projects = {{ {repr(project)}: {repr(doxygenxmlpath)} }}\n")

    def sphinx_transfer(self, srcpath, outpath, heading):
        project, tmpoutpath, doxygenxmlpath = self.__info(outpath)
        super().sphinx_transfer(srcpath, tmpoutpath, heading)
        with open(f"{tmpoutpath}/_doxygen_conf", 'w') as f:
            for k, v in {**self.__doxygenopts, 'STRIP_FROM_PATH': tmpoutpath}.items():
                f.write(f"{k} = {v}\n")
            f.write("EXTENSION_MAPPING += js=JavaScript")
        subprocess.check_call(['doxygen', '_doxygen_conf'], cwd=tmpoutpath)
        subprocess.check_call(['breathe-apidoc', '-o', outpath, '-p', os.path.basename(outpath),
                               '-g', 'class,interface,struct,union,namespace,group', doxygenxmlpath])
        with open(f'{outpath}.rst', 'w') as f:
            f.write(f"""{docgen.heading(heading)}
.. toctree::
   :glob:

   {os.path.basename(outpath)}/*
""")


class CppApiReferenceLanguage(AbstractDoxygenSupportedApiReferenceLanguage):
    """
    C++ language support for api references.
    """

    def __init__(self, **kwargs):
        super().__init__(file_patterns="*.c *.cc *.cxx *.cpp *.c++ *.ii *.ixx *.ipp *.i++ *.inl *.idl *.ddl *.h *.hh"
                                       " *.hxx *.hpp *.h++", **kwargs)


class JavaScriptApiReferenceLanguage(AbstractApiReferenceLanguage):
    """
    JavaScript language support for api references.
    """

    def sphinx_config(self, rootpath, outpath):
        anise.utils.basic.verify_tool_installed('jsdoc')
        return (f"extensions.append('sphinx_js')\n"
                f"js_source_path = {repr(outpath)}\n")

    def __jsfiles(self, df: str) -> List[str]:
        return glob.glob(f"{df}/**/*.js", recursive=True)

    def __scanjsfile(self, f: str) -> str:
        res = ""
        with codecs.getwriter('utf-8')(tempfile.TemporaryFile(mode='w+b')) as ftmp:
            p = subprocess.Popen(['jsdoc', '-X', '-r', f], stdout=ftmp)
            p.wait()
            ftmp.seek(0)
            for jt in json.loads(ftmp.read()):
                name = jt.get('name', "")
                if (not name.startswith("__")) and (jt.get('scope', "") == 'global') and \
                        (not jt.get('undocumented', False)):
                    if jt['kind'] == 'class':
                        res += (f".. js:autoclass:: {name}\n"
                                f"   :members:\n\n")
                    elif jt['kind'] == 'function':
                        res += f".. js:autofunction:: {name}\n\n"
        return res

    def sphinx_transfer(self, srcpath, outpath, heading):
        super().sphinx_transfer(srcpath, outpath, heading)
        cnt = docgen.heading(heading)
        for jsfile in self.__jsfiles(outpath):
            cnt += self.__scanjsfile(jsfile)
        with open(f'{outpath}.rst', 'w') as f:
            f.write(cnt)


class AbstractPiece:
    """
    A subdocument. See :py:func:`render`.
    """

    def __init__(self, name: str, rootpath: str, *, sortidx: int,
                 source: Optional[anise.framework.files.Filestructure] = None):
        """
        :param name: The piece name.
        :param rootpath: The root directory or file that contains the piece content.
        :param sortidx: Sort order index.
        """
        self.__rootpath = rootpath
        self.__source = source
        self.__name = name
        self.__sortidx = sortidx

    @property
    def source(self) -> anise.framework.files.Filestructure:
        """
        The source file structure.
        """
        if self.__source is None:
            self.__source = universe.packages.makerawpackage()
        return self.__source

    @property
    def name(self) -> str:
        """
        The piece name.
        """
        return self.__name

    @property
    def rootpath(self) -> str:
        """
        The root directory or file that contains the piece content.
        """
        return self.__rootpath if os.path.isabs(self.__rootpath) else f"{self.source.path}/{self.__rootpath}"

    @property
    def sortidx(self) -> int:
        """
        The sort order index.
        """
        return self.__sortidx

    def copyto(self, path: str) -> None:
        """
        Copies required data files for this piece to a location.

        :param path: The target location (usually inside the intermediate structure given to Sphinx).
        """
        anise.framework.files.Filestructure(path=self.rootpath).dl(f'{path}/{self.name}')

    def sphinx_config(self, path: str) -> str:
        """
        Returns additional Sphinx :file:`conf.py` configuration.

        :param path: Path to the root directory of the Sphinx input.
        """
        return ""


class RstTextsPiece(AbstractPiece):
    """
    A reStructureText formatted file or a directory of such files.
    """

    def __init__(self, name: str, rootpath: str, *, sortidx: int = 0, source=None):
        """
        :param name: The piece name.
        :param rootpath: The root directory or file that contains the piece content.
        :param sortidx: Sort order index.
        """
        super().__init__(name, rootpath, sortidx=sortidx, source=source)

    def copyto(self, path):
        if os.path.isfile(self.rootpath):
            # also with .rst suffix, so we can include it, but also use it in tocs, without noting '.rst'
            anise.framework.files.Filestructure(path=self.rootpath).dl(f'{path}/{self.name}.rst')
        anise.framework.files.Filestructure(path=self.rootpath).dl(f'{path}/{self.name}')


class ApiReferencePiece(AbstractPiece):
    """
    An api reference.
    """

    def __init__(self, name: str, rootpath: str, language: AbstractApiReferenceLanguage, *,
                 heading: Optional[str] = None, sortidx: int = 100_000, source=None):
        """
        :param name: The piece name.
        :param rootpath: The root directory or file that contains the piece content.
        :param language: The programming language related to this api reference.
        :param heading: The api reference heading.
        :param sortidx: Sort order index.
        """
        super().__init__(name, rootpath, sortidx=sortidx, source=source)
        self.__language = language
        self.__heading = heading or "API reference"

    def copyto(self, path):
        self.__language.sphinx_transfer(self.rootpath, f'{path}/{self.name}', self.__heading)

    def sphinx_config(self, path):
        return self.__language.sphinx_config(self.rootpath, f'{path}/{self.name}')


class AbstractOutputSpec:
    """
    Base class for documentation output specifications. See :py:func:`render`.
    """

    def __init__(self, path: str):
        """
        :param path: The output path.
        """
        self.path = path

    def sphinx_formatname(self) -> str:
        """
        Returns the Sphinx format name.
        """
        raise NotImplementedError()

    def sphinx_config(self, path: str) -> str:
        """
        Returns additional Sphinx :file:`conf.py` configuration.

        :param path: Path to the root directory of the Sphinx input.
        """
        return ""

    def resultpath(self, projectname: str, buildpath: str) -> str:
        """
        Returns the path to the actual result file/directory, relative to :samp:`buildpath`.

        :param projectname: The project name (also document header).
        :param buildpath: The Sphinx build output path.
        """
        raise NotImplementedError()


class HtmlOutputSpec(AbstractOutputSpec):
    """
    HTML documentation output.
    """

    def __init__(self, path: str, *, theme: Optional[str] = None, homepagemode: bool = False,
                 title: Optional[str] = None, short_title: Optional[str] = None, short_desc: Optional[str] = None,
                 masterlink: Optional[str] = None):
        """
        :param path: The output path.
        :param theme: The sphinx html theme name.
        :param homepagemode: If to render output for a homepage with slight different stylings and behaviors.
        :param title: The html title.
        :param short_title: The short html title.
        :param short_desc: The short description. Ignored by most themes.
        :param masterlink: Url that overrides the target of the main heading (which is also a link).
        """
        super().__init__(path)
        self.__theme = theme or "alabaster"
        self.__homepagemode = homepagemode
        self.__title = title
        self.__short_title = short_title
        self.__short_desc = short_desc or ''
        self.__masterlink = masterlink

    def sphinx_formatname(self):
        return "html"

    def sphinx_config(self, path):
        res = f"html_theme = {repr(self.__theme)}\n"
        ato = dict(sidebartreesmall=not self.__homepagemode, menusectionindicator=not self.__homepagemode,
                   shortdesc=self.__short_desc)
        if self.__masterlink:
            ato['masterlink'] = self.__masterlink
        if self.__short_title or self.__title:
            res += f"html_short_title = {repr(self.__short_title or self.__title)}\n"
        if self.__title:
            res += f"html_title = {repr(self.__title)}\n"
        if self.__homepagemode:
            ato.update(sidebarhidelvl1=True, headhidelvl1=True, sidebarhidelvl3up=True, shorthtmltitle=True)
        res += (f"html_theme_options.update(json.loads(base64.b64decode("
                f"{base64.b64encode(json.dumps(ato).encode())})))")
        return res

    def resultpath(self, projectname, buildpath):
        return "html"


class PdfOutputSpec(AbstractOutputSpec):
    """
    PDF documentation output.
    """

    def sphinx_formatname(self):
        return "latexpdf"

    def resultpath(self, projectname, buildpath):
        for f in os.listdir(f"{buildpath}/latex"):
            if f.endswith(".pdf"):
                return f"latex/{f}"
        raise anise.framework.exceptions.InternalError("No pdf found in build output.")


class PlaintextOutputSpec(AbstractOutputSpec):
    """
    Plaintext documentation output.
    """

    def sphinx_formatname(self):
        return "text"

    def resultpath(self, projectname, buildpath):
        return "text"


class RstGenerator:
    """
    Generator for some documentation source parts.
    """

    @staticmethod
    def heading(s: str, *, variant: str = '=', sub: bool = False, anchor: Optional[str] = None) -> str:
        """
        Generates documentation source for a section heading.

        :param s: The heading text.
        :param variant: What heading variant to use. See rst syntax for details.
        :param sub: If to use the sub-variant of :samp:`variant` headings, i.e. with just the symbol line *below* the
                    heading text. See rst syntax for details.
        :param anchor: Anchor name for usage by rst :samp:`ref`s.
        """
        ll = (variant * len(s))
        return "\n\n" + (f".. _{anchor}:\n\n" if anchor else '') + ('' if sub else f'{ll}\n') + f'{s}\n{ll}\n\n'


docgen = RstGenerator


def render(pieces: List[AbstractPiece], outputspec: AbstractOutputSpec,
           projectname: str, *, author: str, heading: Optional[str] = None, mainbody: Optional[str] = None,
           basecolor: Optional[Tuple[float, float, float]] = None, mediadirs: Optional[List[str]] = None,
           exclude_in_toc: Optional[List[str]] = None, include_in_toc: Optional[List[str]] = None,
           shortsnippets: Optional[Dict[str, str]] = None, backgroundimage: Optional[str] = None,
           logoimage: Optional[str] = None) -> None:
    """
    Generate documentation output for some documentation pieces and an output specification.

    :param pieces: Documentation pieces to include.
    :param outputspec: The output specification.
    :param projectname: The project name, used as (part of) top-level title.
    :param author: The project author(s).
    :param heading: The main heading text.
    :param mainbody: reStructuredText source for the main body.
    :param basecolor: The base color (used for styling by some kinds of outputs).
    :param mediadirs: Additional media source paths.
    :param exclude_in_toc: List of piece names to exclude for the table of contents. Default: Exclude none.
    :param include_in_toc: List of piece names to include for the table of contents. Default: Include all.
    :param shortsnippets: Substitution snippets, used by :samp:`|a_foo|` inside document sources.
    :param backgroundimage: Absolute path to the background image (will be copied).
    :param logoimage: Absolute path to the logo image (will be copied).
    """
    anise.utils.basic.verify_tool_installed("sphinx-build")
    if backgroundimage is None:
        backgroundimage = universe.base.getbackgroundimage()
    if logoimage is None:
        logoimage = universe.base.getlogoimage()
    shortsnippets = {**universe.docrender.snippets.getall(), **(shortsnippets or {})}
    mediadirs = (mediadirs or []) + universe.docrender.mediadirs.get_mediadirs()
    psubstitutions = "\n".join([f".. |{k}| replace:: {v}" for k, v in shortsnippets.items()])
    with anise.framework.files.TempDir() as tdir:
        tf = tdir.path
        sconfig = []
        sls = ""
        for mediadir in mediadirs or []:
            for mf in os.listdir(mediadir):
                anise.framework.files.Filestructure(f"{mediadir}/{mf}").dl(tf)
        for piece in pieces:
            piece.copyto(tf)
            sconfig.append(piece.sphinx_config(tf))
            if piece.name not in (exclude_in_toc or []) and (include_in_toc is None or (piece.name in include_in_toc)):
                sls += f"   {piece.name}\n"
        sconfig.append(outputspec.sphinx_config(tf))
        if logoimage:
            anise.framework.files.Filestructure(f"{universe.specialpaths.projectdir}/{logoimage}").dl(f"{tf}/logo.png")
            sconfig.append(f"html_logo = 'logo.png'")
        if backgroundimage:
            anise.framework.files.Filestructure(f"{universe.specialpaths.projectdir}/{backgroundimage}").dl(
                f"{tf}/__anise_background.png")
            sconfig.append(f"html_theme_options['bgimage'] = '__anise_background.png'\n"
                           f"html_static_path.append('__anise_background.png')\n")
        html_theme_options = {}
        basecolor = basecolor or (.1, .2, .3)
        for bcaa in range(10):
            for bcab in range(10):
                for bcac in range(20):
                    sbcac = "abcdefghijklmnopqrst"[bcac]
                    r, g, b = Internals.scalecolor(basecolor, (bcab+1)/10, (bcac+1)/10)
                    html_theme_options[f'brandingcolor_{sbcac}{bcaa}{bcab}'] = (f"rgb({r*255},{g*255},{b*255},"
                                                                                f"{(bcaa+1)/10})")
        sconfig.append(f"html_static_path.append({repr(universe.specialpaths.anisedatadir)}+'/icons/docrender')")
        if universe.base.version:
            sconfig.append(f'version = {repr(str(universe.base.version))}')
        fullversion = universe.packages.getversion()
        if fullversion:
            sconfig.append(f'release = {repr(str(fullversion))}')
        ssconfig = "\n".join(sconfig)
        with open(f'{tf}/conf.py', 'w') as f:
            f.write(f"""
import base64, json, sys
tf = {repr(tf)}
sys.path.insert(0, tf)
project = {repr(projectname)}
author = {repr(author)}
rst_epilog = '''
{psubstitutions}
'''
nitpicky = True
extensions = ['sphinx.ext.autosummary', 'sphinx.ext.inheritance_diagram']
copyright = "{datetime.datetime.now().year}, " + author
exclude_patterns = ['.directory', 'Thumbs.db', '.DS_Store', '.git', '.svn', '.piget', '.idea']
autoclass_content= 'both'
autodoc_typehints = 'description'
html_show_sphinx = False
html_static_path = []
html_theme_path = [{repr(f'{os.path.dirname(__file__)}/docrender_/__static')}]
html_theme_options = json.loads(base64.b64decode({base64.b64encode(json.dumps(html_theme_options).encode())}))
html_sidebars = {{ '**': ['globaltoc.html', 'searchbox.html'] }}
template_ = ''
latex_elements = {{ 'preamble': r'\\usepackage{{enumitem}}\\setlistdepth{{99}}' }}
{ssconfig}
""")
        with open(f'{tf}/index.rst', 'w') as f:
            f.write(f"""{docgen.heading(heading, variant='_') if heading else ''}
.. toctree::
   :hidden:

{sls}

{mainbody}
""")
        with anise.framework.files.TempDir() as tbdir:
            anise.framework.files.Filestructure(f"{universe.specialpaths.anisedatadir}/icons").dl(f"{tf}/aniseicons")
            try:
                subprocess.check_call(["sphinx-build", '-M', outputspec.sphinx_formatname(), tf, tbdir.path],
                                      stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError:
                anise.utils.logging.logerror(f"sphinx-build failed")
                raise
            tbdir.dl(outputspec.path, subpath=outputspec.resultpath(projectname, tbdir.path))
