# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for some basic filesystem operations.
"""

from typing import *
import os
import re
import shutil

import anise.framework.files


def filterbyfunction(acceptorfunction: Callable[[str], bool],
                     source: anise.framework.files.Filestructure) -> anise.framework.files.Filestructure:
    """
    Filters a filestructure by a path acceptor function.

    :param acceptorfunction: A function returning a bool for a relative file path.
    :param source: The source file structure.
    """
    result = anise.framework.files.TempDir()
    dirs = [(source.path, "/")]
    while len(dirs) > 0:
        fulladir, reladir = dirs.pop()
        for afile in os.listdir(fulladir):
            fullafile = fulladir + "/" + afile
            relafile = reladir + "/" + afile
            tgtafile = result.path + "/" + relafile
            if acceptorfunction(relafile):
                if os.path.isdir(fullafile):
                    os.mkdir(tgtafile)
                    dirs.append((fullafile, relafile))
                else:
                    shutil.copy2(fullafile, tgtafile)
    return result


def filterbyextension(source: anise.framework.files.Filestructure, exclude: Optional[List[str]] = None,
                      include: Optional[List[str]] = None,
                      casesensitive: bool = False) -> anise.framework.files.Filestructure:
    """
    Filters a filestructure filename extension.

    :param source: The source file structure.
    :param exclude: A list of extensions to filer away.
    :param include: A list of extensions to not filter away.
    :param casesensitive: If the check is to be applied in a case sensitive manner.
    """
    def acceptorfunction(path):
        def _check(lst):
            ppath = path if casesensitive else path.lower()
            if lst is not None:
                for x in lst:
                    if not x.startswith("."):
                        x = "." + x
                    if not casesensitive:
                        x = x.lower()
                    if ppath.endswith(x):
                        return True
            return False
        if _check(exclude):
            return False
        if _check(include):
            return True
        return include is None
    return filterbyfunction(acceptorfunction, source)


def filterbyname(source: anise.framework.files.Filestructure, exclude: Optional[List[str]] = None,
                 include: Optional[List[str]] = None,
                 casesensitive: bool = False) -> anise.framework.files.Filestructure:
    """
    Filters a filestructure by the file name.

    :param source: The source file structure.
    :param exclude: A list of file names to filer away.
    :param include: A list of file names to not filter away.
    :param casesensitive: If the check is to be applied in a case sensitive manner.
    """
    def acceptorfunction(path):
        def _check(lst):
            pname = os.path.basename(path if casesensitive else path.lower())
            for x in lst or []:
                if pname == x:
                    return True
            return False
        if _check(exclude):
            return False
        if _check(include):
            return True
        return include is None
    return filterbyfunction(acceptorfunction, source)


def filterbynamepattern(source: anise.framework.files.Filestructure, exclude: Optional[List[str]] = None,
                        include: Optional[List[str]] = None,
                        casesensitive: bool = False) -> anise.framework.files.Filestructure:
    """
    Filters a filestructure by a file name pattern.

    :param source: The source file structure.
    :param exclude: A list of file name patterns to filer away.
    :param include: A list of file name patterns to not filter away.
    :param casesensitive: If the check is to be applied in a case sensitive manner.
    """
    def acceptorfunction(path):
        def _check(lst):
            pname = os.path.basename(path)
            if lst is not None:
                for x in lst:
                    if re.match(x, pname, flags=re.IGNORECASE if not casesensitive else 0):
                        return True
            return False
        if _check(exclude):
            return False
        if _check(include):
            return True
        return include is None
    return filterbyfunction(acceptorfunction, source)


def subtree(source: anise.framework.files.Filestructure, root: str = "",
            copy: bool = False) -> anise.framework.files.Filestructure:
    """
    Gets a subtree of a filestructure.

    :param source: The source file structure.
    :param root: The relative path to the new root directory.
    :param copy: If the file structure is to be copied (instead of using the original tree).
    """
    fpath = source.path + "/" + root
    result = anise.framework.files.Filestructure(fpath)
    if copy:
        result = result.with_modified_rootname(os.path.basename(fpath))
    return result
