# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for generating a project homepage.
"""

from typing import *

import anise.framework.features
import anise.framework.files
import anise.framework.report
import anise.utils

from anise.framework.projects import universe


HOOK_UPLOAD_HOMEPAGE_COLLECTLINKS = anise.framework.features.Hook(
    wizard_functionbody="return [('sourcepath', 'destinationpath')]",
    doc="This hook is called for collecting symlinks to deploy together with the homepage downloads.")


class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.homepage.sections = Internals.Sections()

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _basic_homepage_sections():

        # About
        def my_homepage_section_about(**_):
            return universe.base.longdescription

        universe.homepage.sections.add(1000, "about", "About", my_homepage_section_about)

        # Imprint
        def my_homepage_section_imprint(**b):
            if not hasattr(universe.homepage, "imprint"):
                anise.utils.logging.logwarning("The universe object does not define an imprint!")
                return None
            else:
                result = universe.homepage.imprint.replace("[MAIL]", universe.base.email)
                feedbacksect = universe.homepage.sections.getsectionbykey("feedback")
                if feedbacksect:  # if there is a 'feedback' section, we try to include it here
                    result += universe.docgen.heading(feedbacksect[3], variant=':') + feedbacksect[2](**b)
                    universe.homepage.sections.replacetextfct("feedback", lambda *_, **__: None)

                return result

        universe.homepage.sections.add(8000, "imprint", "Imprint", my_homepage_section_imprint)

        # Feedback
        def my_homepage_section_feedback(**_):
            return (f"If you have funny rants about {universe.name} itself or about some techniques it uses, some"
                    f" constructive feedback, a cool patch or a mysterious problem, feel free to mail it"
                    f" to {universe.base.email}.")

        universe.homepage.sections.add(9000, "feedback", "Feedback", my_homepage_section_feedback)

    class Sections:
        """
        Storage of all homepage sections.
        """

        def __init__(self):
            self.list = []

        def add(self, index: int, key: str, header: str, textfct: Callable[[], str]) -> None:
            """
            Adds a homepage section.

            :param index: A position index number. This controls the ordering of sections.
            :param key: The section name. Also used as heading.
            :param header: The section header text.
            :param textfct: A function returning the section documentation source.
            """
            self.list.append([index, key, textfct, header])

        def getsectionbykey(self, key: str) -> Tuple[int, str, Callable[[], str], str]:
            """
            Gets the section record for a key.
            |infrastructure|

            :param key: The section name.
            """
            for l in self.list:
                if l[1] == key:
                    return l

        def gettextfct(self, key: str) -> Callable[[], str]:
            """
            Returns the generator function for a key, that returns the section documentation source.

            :param key: The section name.
            """
            r = self.getsectionbykey(key)
            if r:
                return r[2]

        def replacetextfct(self, key: str, textfct: Callable[[], str]) -> None:
            """
            Replaces a generator function for a key.

            :param key: The section name.
            :param textfct: The new generator function.
            """
            r = self.getsectionbykey(key)
            if r:
                # noinspection PyTupleItemAssignment,PyUnresolvedReferences
                r[2] = textfct


def makehomepage() -> anise.framework.files.Filestructure:
    """
    Creates the project homepage from the registered sections.
    """
    with universe.executionscope.newsubscope("Creating the project homepage"):
        homepage = anise.framework.files.TempDir(dirname=f"{universe.name}-homepage")
        if not getattr(universe, "omitsyncvcs", False):
            universe.versioncontrol.current.syncversioncontrolsystem(
                commitmessage="automated commit for make homepage")
        anise.utils.basic.writetofile(f"{homepage.path}/license", universe.base.license.getlicensetext())
        sections = []
        rawsectionlist = sorted(universe.homepage.sections.list)
        rawsection2scope = []
        for rs in rawsectionlist:
            rawsection2scope.append(universe.executionscope.newsubscope(f"Generating section '{rs[1]}'"))
        topscope = anise.framework.report.topscope()
        for i, section in enumerate(rawsectionlist):
            topscope.progress = anise.framework.report.ExecutionScope.ProgressDeterminate(float(i)/len(rawsectionlist))
            with rawsection2scope[i]:
                key = section[1]
                content = section[2](homepage=homepage)
                header = section[3]
                if content is not None:
                    sections.append((key, content, header))
        universe.documentation.Internals.generate_homepage(head1=universe.name, short_desc=universe.base.summary,
                                                           targetdir=homepage.path, sections=sections)
        return homepage
