# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for creation of Android distribution packages.
"""

import os
import xml.dom.minidom

import anise.framework.exceptions
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


class Internals:

    @staticmethod
    def buildandroidpackage_with_ant(*, projectroot: str, name: str, version: str, targetname: str,
                                     androidpath: str) -> anise.framework.files.Filestructure:
        """
        Builds an Android package with Ant.

        :param projectroot: The path to the Android package.
        :param name: The project name.
        :param version: The project version.
        :param targetname: The Ant target name that shall be built.
        :param androidpath: Path to the Android sdk's :samp:`android` tool.
        """
        with anise.utils.basic.ChDir(projectroot):
            anise.utils.basic.call(androidpath, "update", "project", "-p", ".",
                                   raise_on_errors="unable to call android sdk")
            anise.utils.basic.call("ant", targetname, raise_on_errors="errors while creating android package")
        resfile = None
        for f in os.listdir(f"{projectroot}/bin"):
            if f.endswith(".apk") and not f.endswith("-unaligned.apk"):
                if resfile:
                    raise anise.framework.exceptions.UnexpectedSituationError(f"More than one .apk file found"
                                                                              f" in {projectroot}/bin")
                else:
                    resfile = f
        if resfile is None:
            raise anise.framework.exceptions.UnexpectedSituationError(f"No .apk file found in {projectroot}/bin")
        destresfile = f"{name}-{version}.apk"
        os.rename(f"{projectroot}/bin/{resfile}", f"{projectroot}/{destresfile}")
        return anise.framework.files.Filestructure(f"{projectroot}/{destresfile}")


def package_with_ant(androidpath: str, source: anise.framework.files.Filestructure, relpath: str = "",
                     targetname: str = "debug") -> anise.framework.files.Filestructure:
    """
    Builds an Android package with Ant.

    :param androidpath: Full path to the :samp:`android` tool (something like :file:`.../android-sdk/tools/android`).
    :param source: The source file structure.
    :param relpath: The root path of the android project, relative to `source`s root.
    :param targetname: Name of the Ant target to build.
    """
    with universe.executionscope.newsubscope("Creating Android installation package"):
        return Internals.buildandroidpackage_with_ant(
            name=universe.name, version=universe.packages.getversion(), projectroot=f"{source.path}/{relpath}",
            targetname=targetname, androidpath=androidpath)


def inject_projectdata_to_androidmanifest(outfile: str) -> None:
    """
    Updates some project data (like version number) in the :file:`AndroidManifest.xml`.

    :param outfile: The directory that holds the :file:`AndroidManifest.xml`. Can be absolute or relative to the Anise
                    project directory.
    """
    if outfile.startswith("/"):
        manifest = outfile + "/AndroidManifest.xml"
    else:
        manifest = universe.specialpaths.projectdir + "/" + outfile + "/AndroidManifest.xml"
    version = universe.packages.getversion()
    _version = version.split(".")
    _version2 = []
    if len(_version) > 1:
        _version2.append(_version[0])
    if len(_version) > 2:
        _version2.append(_version[1])
    _version2.append(_version[-1])
    _version2 = ["00000"+x for x in _version2]
    versioncode = int(_version2[0][-2:] + _version2[1][-2:] + _version2[2][-5:])
    doc = xml.dom.minidom.parse(manifest)
    doc.documentElement.attributes["android:versionName"] = version
    doc.documentElement.attributes["android:versionCode"] = str(versioncode)
    with open(manifest, "w") as f:
        doc.writexml(f, encoding="utf-8")
