# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for creation of Debian distribution packages.
"""

from typing import *
import datetime
import gzip
import os
import math
import shutil

import anise.framework.exceptions
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


class Internals:

    @staticmethod
    def builddebpackage(*, name: str, version: str, licensename: str, authors: List[str], email: str, summary: str,
                        description: str, section: str, architecture: str, websiteurl: str,
                        menuentries: List['MenuEntry'], executablelinks: Dict[str, str],
                        requireddependencies: List[str], stronglyrecommendeddependencies: List[str],
                        suggesteddependencies: List[str], services: List['ServiceDescription'], rawpkgpath: str,
                        projectfsroot: str, prerm: str = "", postinst: str = "") -> anise.framework.files.Filestructure:
        """
        Builds a Debian package (:samp:`.deb`).

        :param name: The package name.
        :param version: The package version string.
        :param licensename: The name of the project license.
        :param authors: Project authors.
        :param email: Project email address.
        :param summary: Short project description.
        :param description: Longer project description.
        :param section: The Debian section name.
        :param architecture: The Debian architecture name.
        :param websiteurl: The project homepage url.
        :param menuentries: Menu entries to register.
        :param executablelinks: Symlinks to executables to be created.
        :param requireddependencies: List of required dependencies.
        :param stronglyrecommendeddependencies: List of dependencies that are not strongly required but very typical.
        :param suggesteddependencies: List of optional dependencies.
        :param services: List of services to be registered.
        :param rawpkgpath: The raw source directory; can be damaged.
        :param projectfsroot: The project source directory path.
        :param prerm: Bash code snippet executed before the package is removed on a target machine.
        :param postinst: Bash code snippted executed after the package is installed on a target machine.
        """
        if anise.utils.basic.call("fakeroot -v", shell=True)[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'fakeroot' seems to be unavailable")
        if anise.utils.basic.call("dpkg --version", shell=True)[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'dpkg' seems to be unavailable")
        sdescription = "\n".join([" " + x for x in description.split("\n") if x.strip() != ""])
        conffiles = []
        sdepends = ("\nDepends: " + (", ".join(requireddependencies))) if len(requireddependencies) > 0 else ""
        srecommends = ("\nRecommends: " + (", ".join(stronglyrecommendeddependencies))) \
            if len(stronglyrecommendeddependencies) > 0 else ""
        ssuggests = ("\nSuggests: " + (", ".join(suggesteddependencies))) if len(suggesteddependencies) > 0 else ""
        workdir = anise.framework.files.TempDir().path
        iworkdir = f"{workdir}/pkgtemp"
        packagerootname = f"{name}-{version}-{architecture}"
        packageroot = f"{iworkdir}/fs/{packagerootname}"
        placedoc = f"{packageroot}/usr/share/doc/{name}"
        placepixmaps = f"{packageroot}/usr/share/pixmaps"
        placedebian = f"{packageroot}/DEBIAN"
        os.makedirs(iworkdir)
        os.makedirs(placedoc)
        if not rawpkgpath.endswith("/"):
            rawpkgpath += "/"
        _files = [rawpkgpath]
        while len(_files) > 0:
            f = _files.pop()
            if os.path.isdir(f):
                for x in os.listdir(f):
                    _files.append(f"{f}/{x}")
            else:
                ff = f[len(rawpkgpath) + 1:]
                gg = f"{packageroot}/{ff}"
                dd = os.path.dirname(gg)
                if not os.path.exists(dd):
                    os.makedirs(dd)
                shutil.copy2(f, gg)
        if not os.path.isdir(f"{packageroot}/usr/bin"):
            os.makedirs(f"{packageroot}/usr/bin")
        for executablelink in executablelinks.keys():
            os.symlink(executablelinks[executablelink], f"{packageroot}/usr/bin/{executablelink}")
        licensetext = f"""
Format: http://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: {name}
Source: {websiteurl}
Upstream-Contact: The {name} Team <{email}>

Files: *
Copyright: {datetime.datetime.now().strftime("%Y")} {authors}
License: {licensename}
"""[1:]
        with open(f"{placedoc}/copyright", "w") as f:
            f.write(licensetext)
        changelog = f"""
{name} {version} unstable; urgency=low

* New upstream release.
* See website for details.

-- The {name} Team <{email}>  Mon, 14 Jan 2013 13:37:00 +0000
"""[1:]
        cchangelog = gzip.compress(changelog.encode("utf-8"))
        with open(f"{placedoc}/changelog.gz", "wb") as f:
            f.write(cchangelog)
        os.makedirs(placepixmaps)
        os.makedirs(f"{packageroot}/usr/share/menu")
        os.makedirs(f"{packageroot}/usr/share/applications")
        for menuentry in menuentries:
            sdebiancategory = menuentry.category[0]
            sfreedesktopcategory = menuentry.category[1]
            ename = menuentry.name
            etitle = menuentry.title
            cmd = menuentry.command
            cmd2 = cmd.replace('"', '\\"')
            iconsrc = f"{projectfsroot}/{menuentry.icon}"
            iconfname = f"{name}.{ename}.png"
            shutil.copyfile(iconsrc, f"{placepixmaps}/{iconfname}")
            sisterminal = "false" if menuentry.gui else "true"
            sneeds = "X11" if menuentry.gui else "text"
            with open(f"{packageroot}/usr/share/menu/{ename}", "w") as f:
                f.write(f"""
?package({ename}): \\
  command="{cmd2}" \\
  needs="{sneeds}" \\
  section="{sdebiancategory}" \\
  title="{etitle}" \\
  icon="/usr/share/pixmaps/{iconfname}"
"""[1:])
            with open(f"{packageroot}/usr/share/applications/{ename}.desktop", "w") as f:
                f.write(f"""
[Desktop Entry]
Name={etitle}
Exec={cmd}
Terminal={sisterminal}
Type=Application
Categories={sfreedesktopcategory};
Icon=/usr/share/pixmaps/{iconfname}
"""[1:])
        os.makedirs(f"{packageroot}/etc/init")
        startservicescall = ""
        stopservicescall = ""
        for service in services:
            servicename = service.name
            servicecommand = service.command
            with open(f"{packageroot}/etc/init/{servicename}.conf", "w") as f:  # TODO systemd services
                f.write(f"""
# {name} - {name} job file

description "{name} service '{servicename}'"
author "The {name} Team <{email}>"

start on runlevel [2345]

stop on runlevel [016]

exec {servicecommand}
"""[1:])
            startservicescall += f"service {servicename} start\n"
            stopservicescall += f"service {servicename} stop &>/dev/null\n"
            conffiles.append(f"/etc/init/{servicename}.conf")
        size = 0
        for dirpath, dirnames, filenames in os.walk(packageroot):
            for f in filenames:
                ff = f"{dirpath}/{f}"
                if not os.path.islink(ff):
                    size += os.path.getsize(ff)
        size = int(math.ceil(size / 1024.))
        os.makedirs(placedebian)
        # license
        with open(f"{placedebian}/copyright", "w") as f:
            f.write(licensetext)
        # prerm
        with open(f"{placedebian}/prerm", "w") as f:
            f.write(f"""
{stopservicescall}
set -e
{prerm}
"""[1:])
        # postinst
        with open(f"{placedebian}/postinst", "w") as f:
            f.write(f"""
#!/bin/bash
set -e
{postinst}
{startservicescall}
if test -x /usr/bin/update-menus; then update-menus; fi
"""[1:])
        # control
        with open(f"{placedebian}/control", "w") as f:
            f.write(f"""
Package: {name}
Version: {version}
Section: {section}
Priority: optional
Architecture: {architecture}{sdepends}{srecommends}{ssuggests}
Installed-Size: {size}
Maintainer: The {name} Team <{email}>
Provides: {name}
Homepage: {websiteurl}
Description: {summary}
{sdescription}
"""[1:])
        # conffiles
        with open(f"{placedebian}/conffiles", "w") as f:
            f.write("\n".join(conffiles))
        anise.utils.basic.call("chmod", "-R", "u+rw,g+r,g-w,o+r,o-w", packageroot)
        anise.utils.basic.call("chmod", "-R", "0755", placedebian)
        anise.utils.basic.call("chmod", "0644", f"{placedebian}/conffiles")
        debfilename = f"{name}_{version}_{architecture}.deb"
        r, o = anise.utils.basic.call(["fakeroot", "dpkg", "-b", packageroot, f"{workdir}/{debfilename}"],
                                      raise_on_errors="unable to create debian package")
        anise.utils.logging.logdebug(o)
        return anise.framework.files.Filestructure(f"{workdir}/{debfilename}")


class Category:
    """
    Enumeration for Debian's categories for software packages.
    """

    ApplicationsAccessibility = ("Applications/Accessibility", "System")
    ApplicationsAmateurradio = ("Applications/Amateur Radio", "Utility")
    ApplicationsDatamanagement = ("Applications/Data Management", "System")
    ApplicationsEditors = ("Applications/Editors", "Utility")
    ApplicationsEducation = ("Applications/Education", "Education")
    ApplicationsEmulators = ("Applications/Emulators", "System")
    ApplicationsFilemanagement = ("Applications/File Management", "System")
    ApplicationsGraphics = ("Applications/Graphics", "Graphics")
    ApplicationsMobiledevices = ("Applications/Mobile Devices", "Utility")
    ApplicationsNetwork = ("Applications/Network", "Network")
    ApplicationsNetworkCommunication = ("Applications/Network/Communication", "Network")
    ApplicationsNetworkFiletransfer = ("Applications/Network/File Transfer", "Network")
    ApplicationsNetworkMonitoring = ("Applications/Network/Monitoring", "Network")
    ApplicationsNetworkWebbrowsing = ("Applications/Network/Web Browsing", "Network")
    ApplicationsNetworkWebnews = ("Applications/Network/Web News", "Network")
    ApplicationsOffice = ("Applications/Office", "Office")
    ApplicationsProgramming = ("Applications/Programming", "Development")
    ApplicationsProjectmanagement = ("Applications/Project Management", "Development")
    ApplicationsScience = ("Applications/Science", "Education")
    ApplicationsScienceAstronomy = ("Applications/Science/Astronomy", "Education")
    ApplicationsScienceBiology = ("Applications/Science/Biology", "Education")
    ApplicationsScienceChemistry = ("Applications/Science/Chemistry", "Education")
    ApplicationsScienceDataanalysis = ("Applications/Science/Data Analysis", "Education")
    ApplicationsScienceElectronics = ("Applications/Science/Electronics", "Education")
    ApplicationsScienceEngineering = ("Applications/Science/Engineering", "Education")
    ApplicationsScienceGeoscience = ("Applications/Science/Geoscience", "Education")
    ApplicationsScienceMathematics = ("Applications/Science/Mathematics", "Education")
    ApplicationsScienceMedicine = ("Applications/Science/Medicine", "Education")
    ApplicationsSciencePhysics = ("Applications/Science/Physics", "Education")
    ApplicationsScienceSocial = ("Applications/Science/Social", "Education")
    ApplicationsShells = ("Applications/Shells", "System")
    ApplicationsSounds = ("Applications/Sound", "AudioVideo")
    ApplicationsSystem = ("Applications/System", "System")
    ApplicationsSystemAdministration = ("Applications/System/Administration", "System")
    ApplicationsSystemHardware = ("Applications/System/Hardware", "System")
    ApplicationsSystemLanguageenvironment = ("Applications/System/Language Environment", "System")
    ApplicationsSystemMonitoring = ("Applications/System/Monitoring", "System")
    ApplicationsSystemPackagemanagement = ("Applications/System/Package Management", "System")
    ApplicationsSystemSecurity = ("Applications/System/Security", "System")
    ApplicationsTerminalemulators = ("Applications/Terminal Emulators", "System")
    ApplicationsText = ("Applications/Text", "Utility")
    ApplicationsTvandradio = ("Applications/TV and Radio", "AudioVideo")
    ApplicationsViewers = ("Applications/Viewers", "Utility")
    ApplicationsVideo = ("Applications/Video", "AudioVideo")
    ApplicationsWebdevelopment = ("Applications/Web Development", "Development")
    GamesAction = ("Games/Action", "Game")
    GamesAdventure = ("Games/Adventure", "Game")
    GamesBlocks = ("Games/Blocks", "Game")
    GamesBoard = ("Games/Board", "Game")
    GamesCard = ("Games/Card", "Game")
    GamesPuzzles = ("Games/Puzzles", "Game")
    GamesSimulation = ("Games/Simulation", "Game")
    GamesStrategy = ("Games/Strategy", "Game")
    GamesTools = ("Games/Tools", "Game")
    GamesToys = ("Games/Toys", "Game")
    Help = ("Help", "System")
    ScreenSaving = ("Screen/Saving", "System")
    ScreenLocking = ("Screen/Locking", "System")


class ServiceDescription:
    """
    Description for Debian services to be included in a package.
    """

    def __init__(self, name: str, command: str):
        """
        :param name: The display name.
        :param command: The command to be executed.
        """
        self.name = name
        self.command = command


class MenuEntry:
    """
    Specification for one menu entry added by a Debian installation package.
    """

    def __init__(self, *, name: str, title: str, category: Tuple, command: str, gui: bool, icon: str):
        """
        :param name: Internal menu entry name.
        :param title: Menu entry Title.
        :param category: Menu entry category. One of :py:class:`Category`.
        :param command: Menu entry command.
        :param gui: If the command opens a gui (instead of a terminal application).
        :param icon: The path to a menu entry icon.
        """
        self.name = name
        self.title = title
        self.category = category
        self.command = command
        self.gui = gui
        self.icon = icon


# noinspection PyShadowingNames
def debpackage(source: anise.framework.files.Filestructure,
               dependencies: Optional[List['dependencies.Dependency']] = None,
               executablelinks: Optional[Dict[str, str]] = None, menuentries: Optional[List[MenuEntry]] = None,
               services: Optional[List[ServiceDescription]] = None, prerm: str = "", postinst: str = "",
               architecture: str = "all") -> anise.framework.files.Filestructure:
    """
    Builds a Debian package.

    :param source: A source file structure.
    :param dependencies: A list of Debian dependencies for installation on target machines.
    :param executablelinks: A dict for links to executables for deployment on target machines.
    :param menuentries: A list of menu entries for creation on target machines.
    :param services: A list of services for deployment on target machines.
    :param prerm: Piece of bash script executed before removal of the package on target machines.
    :param postinst: Piece of bash script executed after installation of the package on the target machines.
    :param architecture: The target architecture name.
    """
    with universe.executionscope.newsubscope("Creating debian package"):
        srcsrcfs = universe.packages.makerawpackage().mv(universe.name)
        for src, dst in [("README", f"/usr/share/doc/{universe.name}/README"),
                         ("README.pdf", f"/usr/share/doc/{universe.name}/README.pdf")]:
            s = f"{srcsrcfs.path}/{src}"
            d = os.path.abspath(f"{source.path}/{dst}")
            if os.path.isfile(s):
                anise.utils.basic.mkdirs(os.path.dirname(d))
                srcsrcfs.dl(d, subpath=src)
        executablelinks = executablelinks or {}
        menuentries = menuentries or []
        services = services or []
        dependencies = list(dependencies or []) + universe.dependencies.pool.list
        dependencies = [x for x in dependencies if hasattr(x, "debian")]
        reqdependencies = [x for x in dependencies if (x.deptype in [universe.dependencies.Type.Required,
                                                                     universe.dependencies.Type.Included])]
        recdependencies = [x for x in dependencies
                           if (x.deptype == universe.dependencies.Type.Required_HasAlternatives)]
        sugdependencies = [x for x in dependencies
                           if (x.deptype == universe.dependencies.Type.Recommended_HasAlternatives)]
        reqdependencies = [dependency for dependencyitem in reqdependencies for dependency in dependencyitem.debian]
        recdependencies = [dependency for dependencyitem in recdependencies for dependency in dependencyitem.debian]
        sugdependencies = [dependency for dependencyitem in sugdependencies for dependency in dependencyitem.debian]
        return Internals.builddebpackage(
            name=universe.name, version=universe.packages.getversion(),
            licensename=universe.base.license.namedebian, authors=universe.base.authors[0],
            email=universe.base.email, summary=universe.base.summary, description=universe.base.longdescription,
            section="web", menuentries=menuentries, executablelinks=executablelinks,
            requireddependencies=reqdependencies, stronglyrecommendeddependencies=recdependencies,
            suggesteddependencies=sugdependencies, services=services, architecture=architecture,
            websiteurl=universe.base.websiteurl, rawpkgpath=source.path,
            projectfsroot=universe.specialpaths.projectdir, prerm=prerm, postinst=postinst)


def only_programfiles(source):
    """
    Returns a filtered set of files for a Debian program directory (no documentation, license, tests, ...).

    :param source: An anise.framework.files.Filestructure.
    :return: A filtered anise.framework.files.Filestructure.
    """
    srcfs = source.with_modified_rootname("x")
    for remv in ["README", "README.pdf", "LICENSE", "_meta", "_test"]:
        x = f"{srcfs.path}/{remv}"
        if os.path.isfile(x):
            os.unlink(x)
        elif os.path.isdir(x):
            shutil.rmtree(x)
    return srcfs


if TYPE_CHECKING:
    import anise.features.dependencies as dependencies
