# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for creation of oci container images.
"""

from typing import *
import shlex

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.packages.oci.fullnamepattern = getattr(universe.packages.oci, "fullnamepattern", "{name}")

    @staticmethod
    def call(cmd: List[str], raise_on_errors: Optional[str] = "unable to run buildah") -> str:
        r, o = anise.utils.basic.call(cmd, raise_on_errors=raise_on_errors)
        anise.utils.logging.logdebug(o)
        return o


class PortDescription:
    """
    Description for Oci network ports to be added to the container image.
    """

    def __init__(self, port: int):
        """
        :param port: The port number as available inside the container.
        """
        self.port = int(port)


class EnvironmentVariableDescription:
    """
    Description for Oci environment variables to be added to the container image.
    """

    def __init__(self, name: str, value: str):
        """
        :param name: The variable name.
        :param value: The value.
        """
        self.name = name
        self.value = value


def ocidebianpackage(debianpackage: anise.framework.files.Filestructure, fullname: Optional[str] = None,
                     ports: Optional[List[PortDescription]] = None,
                     environment_variables: Optional[List[EnvironmentVariableDescription]] = None,
                     execcommand: Optional[List[str]] = None, run: Optional[List[List[str]]] = None,
                     run_late: Optional[List[List[str]]] = None) -> anise.framework.files.Filestructure:
    """
    Builds an oci container image based on Debian.
    
    You need an existing Debian package source.

    :param debianpackage: A file structure containing the debian package.
    :param fullname: The full container name.
    :param ports: A list of PortDescription port descriptions to add to the container image.
    :param environment_variables: A list of EnvironmentVariableDescription descriptions to add to the container image.
    :param execcommand: A list of strings to specify the command to execute inside the container after boot.
    :param run: A list of command-line lists to execute for preparation.
    :param run_late: A list of command-line lists to execute for preparation after installing the package.
    """
    with universe.executionscope.newsubscope("Creating oci container image"):
        if anise.utils.basic.call(["buildah", "--version"])[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'buildah' seems to be unavailable")
        if anise.utils.basic.call(["podman", "--version"])[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'podman' seems to be unavailable")
        sauthors = ", ".join(universe.base.authors)
        projectversion = universe.packages.getversion()
        containerdir = anise.framework.files.TempDir().path
        imgname = (fullname or universe.packages.oci.fullnamepattern).format(name=universe.name)
        ports = ports or []
        environment_variables = environment_variables or []
        run = run or []
        run_late = run_late or []
        with anise.utils.basic.ChDir(containerdir):
            cn = Internals.call(["buildah", "from", "docker.io/debian:latest"]).strip()
            ociname = "{name}-{version}.oci.tar".format(name=imgname, version=projectversion)
            Internals.call(["buildah", "copy", cn, debianpackage.path, "/root/pkg.deb"])
            Internals.call(["buildah", "run", cn, "--", "apt-get", "update"])
            for l in run:
                Internals.call(["buildah", "run", cn, "--", *l], raise_on_errors="unable to run 'run' stuff")
            for e in environment_variables:
                Internals.call(["buildah", "config", "--env", f"{e.name}={e.value}", cn])
            Internals.call(["buildah", "run", cn, "--", "/usr/bin/dpkg", "-i", "/root/pkg.deb"],
                           raise_on_errors=None)
            Internals.call(["buildah", "run", cn, "--", "apt-get", "-f", "-y", "install"])
            Internals.call(["buildah", "config", "--author", sauthors, cn])
            for p in ports:
                Internals.call(["buildah", "config", "--port", str(p.port), cn])
            for l in run_late:
                Internals.call(["buildah", "run", cn, "--", *l], raise_on_errors="unable to run 'run_late' stuff")
            if execcommand:
                sexecmd = " ".join([shlex.quote(x) for x in execcommand])
                Internals.call(["buildah", "config", "--entrypoint", sexecmd, cn])
            Internals.call(["buildah", "commit", cn, imgname])
            Internals.call(["podman", "save", "-o", ociname, imgname])
            Internals.call(["buildah", "rmi", "--force", imgname])
            return anise.framework.files.Filestructure(f"{containerdir}/{ociname}")
