# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Features for graphical interfaces that allow making changes to the project description file. This includes a raw
editor, but also a framework for configuration assistants, which guide the user through some configuration steps.

Note: All inner features are internal and not required to directly work with in usual cases. They provide
assistants for configuration of the Anise builtin features. Extending those particular assistants (in contrast to
implementing new ones) is one of the cases you need to work with the inner features here. Doing so typically means
finding a meaningful hook inside and register a handler to it.
"""

from typing import *

import anise.features.diagnostics
import anise.framework.exceptions
import anise.framework.features
import anise.framework.projects

from anise.framework.projects import universe

diagnostics = anise.framework.features.loadfeature("diagnostics").featuremodule
ui = anise.framework.features.loadfeature("ui").featuremodule


#: Hook triggered for collecting all implementations for feature actions, as offered in the console.
HOOK_GET_FEATURE_ACTIONS = diagnostics.HOOK_GET_FEATURE_ACTIONS

#: Hook triggered for collecting custom actions of nodes, for usage in a configuration guide.
HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS = anise.framework.features.Hook(
    hide=True, doc="This hook is called for collecting custom actions of nodes, for usage in a configuration guide.")


class Internals:

    @anise.framework.features.hookhandler(diagnostics.HOOK_GET_FEATURE_ACTIONS, provides=["raweditor"])
    class EditProjectdescApplication(diagnostics.FeatureAction):
        """
        The raw mode editor. It offers a configuration interface that corresponds to the file format, but also
        custom actions for some nodes, which may be helpful for configuration.
        """

        def __init__(self):
            super().__init__("Raw Mode Editor", "Helps to edit the project description in a way that directly"
                                                " corresponds to the file format.", forfeature="")
            self.ais = None

        @staticmethod
        def editnode(ais: anise.framework.projects.IntermediateStructure,
                     node: anise.framework.projects.IntermediateStructure.Node) -> None:
            """
            Runs an editor for a particular node.

            :param ais: A project structure.
            :param node: A project structure node.
            """
            epa = Internals.EditProjectdescApplication()
            epa.init(ais)
            epa.changeentry(node)

        def summarizeentry(self, e: str) -> str:
            def lines(s):
                return 1 + len([c for c in s if c == "\n"])
            maxlen = 600
            maxlines = 5
            x = str(e)
            if (len(x) <= maxlen) and (lines(x) <= maxlines):
                return x
            else:
                c1 = int(maxlen / 2)
                c2 = len(x) - (maxlen - c1)
                while lines(x[:c1]) > int(maxlines/2):
                    c1 = x.rfind("\n", 0, c1)
                while lines(x[c2:]) > maxlines - int(maxlines/2) - 1:
                    c2 = x.find("\n", c2) + 1
                return x[:c1] + "\n . . .\n" + x[c2:]

        def _store(self) -> None:
            self.ais.save()

        def init(self, ais: Optional[anise.framework.projects.IntermediateStructure] = None) -> None:
            """
            Initializes the editor.

            This is required if you do not want to execute the editor but only use internal parts of it.

            :param ais: A project structure.
            """
            self.ais = ais or anise.framework.projects.IntermediateStructure.fromfile(universe.specialpaths.projectfile)

        def execute(self) -> None:
            """
            Executes the editor.
            """
            self.init()
            self.mainmenu()

        def addletentry(self):
            key = universe.ui.userfeedback.input("Please choose a key for the new let.", "feature.foo")
            if key is not None:
                n = anise.framework.projects.IntermediateStructure.Let(key)
                self.ais.addentry(n)
                self._store()

        def changevalvalue(self, entry: anise.framework.projects.IntermediateStructure.Value):
            result = False
            if entry.vtype in ["str", "int", "float"]:
                success = False
                while not success:
                    if entry.vtype == "str":
                        nv = universe.ui.userfeedback.multilineinput(
                            "Please choose a new value for '{0}'.".format(entry.key or "list entry"), entry.value)
                    else:
                        nv = universe.ui.userfeedback.input(
                            "Please choose a new value for '{0}'.".format(entry.key or "list entry"), entry.value)
                    if nv is None:
                        success = True
                    else:
                        try:
                            # noinspection PyProtectedMember
                            entry.value = nv
                            success = True
                            result = True
                        except ValueError:
                            universe.ui.userfeedback.messageok(
                                "Please enter a valid '{0}' value.".format(entry.vtype))
            elif entry.vtype == "objectref":
                nv = universe.ui.userfeedback.multilineinput(
                    "Please choose a new reference string for:\n" + self.summarizeentry(entry), entry.oref)
                if nv is not None:
                    entry.oref = nv
                    result = True
            elif entry.vtype == "bool":
                nv = universe.ui.userfeedback.message(
                    "Please choose a new value.".format(entry.value), ["True", "False", "Cancel"])
                if nv == 0:
                    entry.value = "1"
                    result = True
                elif nv == 1:
                    entry.value = "0"
                    result = True
            return result

        def addvalentry(self, key=None, getkeyfromargs=False):
            ll = ["str", "int", "bool", "float", "list", "dict", "objectref", "another type"]
            i = universe.ui.userfeedback.choice("Please choose a data type for the new value.", ll)
            if i is not None:
                vtype = ll[i]
                if vtype == ll[-1]:
                    vtype = universe.ui.userfeedback.input("Please choose a data type for the new value.")
                if vtype is not None:
                    if not getkeyfromargs:
                        key = universe.ui.userfeedback.input("Please choose a key for the new value.", "foo")
                    if key is not None:
                        prm = {}
                        n = anise.framework.projects.IntermediateStructure.Value(vtype, k=key, **prm)
                        if vtype in ["str", "int", "bool", "float", "objectref"]:
                            store = self.changevalvalue(n)
                        else:
                            store = True
                        if store:
                            return key, n
            return None, None

        def movelistchildentry(self, entryargs, fct_move):
            i1 = universe.ui.userfeedback.choice("Please choose which entry you want to move.", entryargs)
            if i1 is not None:
                i2 = universe.ui.userfeedback.choice("Please choose a new place.", entryargs)
                if i2 is not None:
                    fct_move(i1, i2)

        def changeentry(self, entry: anise.framework.projects.IntermediateStructure.Node, remove_is_unsafe=True):
            while True:
                ll = []
                limages = []
                _icustomactions = {}
                customactions = [x() for x
                                 in universe.gethooks(HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS)]
                for i, customaction in enumerate(customactions):
                    if customaction.visible(self.ais, entry):
                        _icustomactions[len(ll)] = i
                        ll.append(customaction.label)
                        limages.append(customaction.icon)
                _ientryargsactions = {}
                entryargskeys = sorted(entry.args())
                for i, x in enumerate(entryargskeys):
                    if isinstance(entry, anise.framework.projects.IntermediateStructure.Value) \
                            and entry.vtype == "list":
                        fmt = "{0}: {2}"
                    else:
                        fmt = "More for '{1}'"
                    _ientryargsactions[len(ll)] = i
                    ll.append(fmt.format(i, x, entry.getarg(x)))
                    limages.append("userfeedback/find.png")
                allowadd = True
                _iremoveaction = -1
                _ichangevaction = -1
                _ichangetypeaction = -1
                _iaddvalueaction = -1
                _iaddlistentryaction = -1
                _imovelistentryaction = -1
                if isinstance(entry, anise.framework.projects.IntermediateStructure.Value):
                    _ichangetypeaction = len(ll)
                    ll.append("Change type")
                    limages.append("userfeedback/preferences.png")
                    if entry.vtype in ["str", "int", "bool", "float", "objectref"]:
                        _ichangevaction = len(ll)
                        ll.append("Change value")
                        limages.append("userfeedback/preferences.png")
                        allowadd = False
                    elif entry.vtype == "list":
                        _iaddlistentryaction = len(ll)
                        ll.append("Add subvalue")
                        limages.append("userfeedback/add.png")
                        _imovelistentryaction = len(ll)
                        ll.append("Move a subvalue")
                        limages.append("userfeedback/magic.png")
                        allowadd = False
                _iremoveaction = len(ll)
                ll.append("Remove")
                limages.append("userfeedback/remove.png")
                if allowadd:
                    _iaddvalueaction = len(ll)
                    ll.append("Add subvalue")
                    limages.append("userfeedback/add.png")
                imod = universe.ui.userfeedback.choice(
                    "Please choose what to do for:\n\n" + self.summarizeentry(entry), ll, canceltext="Back",
                    choiceimages=limages)
                if imod == _iremoveaction:
                    confirmmsg = "Do you really want to delete this entry?\n\n" + self.summarizeentry(entry)
                    if remove_is_unsafe:
                        confirmmsg = ("Warning: Removing structures in the project description may break the"
                                      " configuration, e.g. leading to a missing parameter of a constructor or"
                                      " function. Please ensure that removing this item is valid before you"
                                      " confirm!\n\n" + confirmmsg)
                    if universe.ui.userfeedback.messageyesno(confirmmsg):
                        entry.remove()
                        self._store()
                        break
                elif imod == _ichangetypeaction:
                    otype = entry.vtype
                    ntype = universe.ui.userfeedback.input(
                        "Warning: Changing the value type makes sense in rare cases and is only allowed if the"
                        " sub-values will match to the new type as well.\n\nPlease choose the new"
                        " type for\n" + self.summarizeentry(entry), otype)
                    if ntype != otype and (ntype is not None):
                        entry.vtype = ntype
                        self._store()
                elif imod == _ichangevaction:
                    if self.changevalvalue(entry):
                        self._store()
                elif imod == _iaddvalueaction:
                    key, node = self.addvalentry()
                    if node:
                        entry.setarg(key, node)
                        self._store()
                elif imod == _iaddlistentryaction:
                    key, node = self.addvalentry(getkeyfromargs=True, key="foo")
                    if node:
                        entry.appendarg(node)
                        self._store()
                elif imod == _imovelistentryaction:
                    def _move(_i, _j):
                        ekeys = sorted(entry.args())
                        jj = _j
                        if _i < _j:
                            jj += 1
                        nkey = (ekeys[jj-1] + "_") if _j > 0 else "i0"
                        o = entry.removearg(ekeys[_i])
                        entry.setarg(nkey, o)
                        # noinspection PyUnresolvedReferences,PyProtectedMember
                        entry._realign_listvalues()
                        self._store()
                    self.movelistchildentry([str(entry.getarg(x)) for x in sorted(entry.args())], _move)
                elif imod in _ientryargsactions:
                    self.changeentry(entry.getarg(entryargskeys[_ientryargsactions[imod]]))
                elif imod in _icustomactions:
                    customactions[_icustomactions[imod]].function(self.ais, entry)
                else:
                    break

        def mainmenu(self):
            while True:
                choicetree = universe.ui.ChoiceTree()
                stack = []
                for e in self.ais.entries():
                    stack.append((choicetree.root, e))
                while len(stack) > 0:
                    node, ent = stack[0]
                    stack = stack[1:]
                    key = "+" if (isinstance(ent.parent, anise.framework.projects.IntermediateStructure.Value)
                                  and ent.parent.vtype == "list") else ent.key
                    if isinstance(ent, anise.framework.projects.IntermediateStructure.Value):
                        prrepr = ent.getprimitiverepr()
                        if prrepr is not None:
                            key = key + " = " + prrepr
                        else:
                            key = key + " (" + ent.vtype + ")"
                    if isinstance(ent, anise.framework.projects.IntermediateStructure.Value):
                        if ent.vtype == "str":
                            image = "pde/vstr.png"
                        elif ent.vtype == "int":
                            image = "pde/vnum.png"
                        elif ent.vtype == "bool":
                            image = "pde/vbool.png"
                        elif ent.vtype == "float":
                            image = "pde/vnum.png"
                        elif ent.vtype == "objectref":
                            image = "pde/voref.png"
                        elif ent.vtype == "list":
                            image = "pde/vlist.png"
                        elif ent.vtype == "dict":
                            image = "pde/vdict.png"
                        else:
                            image = "pde/vother.png"
                    elif isinstance(ent, anise.framework.projects.IntermediateStructure.Let):
                        image = "pde/let.png"
                    else:
                        raise anise.framework.exceptions.ImplementationError("unknown thing: " + str(ent))
                    newnode = node.createchild(key, ent, image=image)
                    prms = {k: ent.getarg(k) for k in ent.args()}
                    for c in sorted(prms.keys()):
                        stack.append((newnode, prms[c]))
                _sa_newval = "New value"
                _sa_newlet = "New let"
                _sa_move = "Move"
                c = universe.ui.userfeedback.treechoice(
                    "Please choose an entry for more or add a new one.", choicetree, canceltext="Back",
                    actions=[_sa_newval, _sa_newlet, _sa_move])
                if c == _sa_newval:
                    key, node = self.addvalentry()
                    if node:
                        self.ais.addentry(node)
                        self._store()
                elif c == _sa_newlet:
                    self.addletentry()
                elif c == _sa_move:
                    def _move(oldpos, newpos):
                        olditem = self.ais.removeentryat(oldpos)
                        self.ais.setentries(self.ais.entries()[0:newpos] + [olditem] + self.ais.entries()[newpos:])
                        self._store()
                    self.movelistchildentry(self.ais.entries(), _move)
                elif c is not None:
                    self.changeentry(c.val)
                else:
                    break


class CustomAction:
    """
    A custom action in a :py:class:`ProjectDescriptionEditorAssistant`. Can either be a global one or working on a
    particular node.
    """

    def __init__(self, label: str, function: Optional[Callable] = None, icon: Optional[str] = None):
        """
        :param label: The label text.
        :param function: The action function (if :samp:`None`: :samp:'self.execute' is used).
        :param icon: Icon name.
        """
        self.label = label
        if function:
            self.function = lambda *a, **b: function(*a, **b)
        else:
            self.function = lambda *a, **b: self.execute(*a, **b)
        self.icon = icon

    # noinspection PyUnusedLocal
    def visible(self, ais: anise.framework.projects.IntermediateStructure,
                node: anise.framework.projects.IntermediateStructure.Node) -> bool:
        """
        Checks if this custom action shall be visible.
        |dooverride_optional|

        :param ais: A project structure.
        :param node: A project structure node.
        """
        return True

    def execute(self, ais: anise.framework.projects.IntermediateStructure,
                node: anise.framework.projects.IntermediateStructure.Node) -> None:
        """
        Executes this custom action.
        |dooverride|

        :param ais: A project structure.
        :param node: A project structure node.
        """
        raise anise.framework.exceptions.NotImplementedError()


class ProjectDescriptionEditorAssistant(anise.features.diagnostics.FeatureAction):
    """
    Interface class for a project description editor assistant.

    Represents one entry in the main list.
    """

    def __init__(self, label: str, description: str, forfeature: str = "", header: Optional[str] = None):
        """
        :param label: The action label text.
        :param description: The action description text.
        :param forfeature: Full name of the feature to associate this action with (e.g. :samp:`'foo.bar'`).
        :param header: Header text for the overview screen.
        """
        super().__init__(label, description, forfeature)
        self.header = header or description

    def execute(self):
        ais = anise.framework.projects.IntermediateStructure.fromfile(universe.specialpaths.projectfile)
        while True:
            c = []
            cimgs = []
            idx2lbl = []
            idx2ent = []
            idx2act = []
            for d in self.getentries(ais):
                try:
                    lbl = self.getentrylabel(ais, d)
                    if lbl is None:
                        raise Exception()
                except Exception:
                    lbl = "unknown: " + str(d)  # only fallback
                    if len(lbl) > 100:
                        lbl = lbl[:96] + " ..."
                    idx2lbl.append(lbl)
                idx2ent.append(d)
                idx2act.append((lambda _d: lambda _ais: self.manageentry(_ais, _d))(d))
                c.append("'" + lbl + "'")
                cimgs.append("pde/vother.png")
            for ca in self.getcustomactions(ais):
                # noinspection PyTypeChecker
                if ca.visible(None, None):
                    c.append(ca.label)
                    cimgs.append(ca.icon)
                    idx2lbl.append(None)
                    idx2ent.append(None)
                    idx2act.append(ca.function)
            ic = universe.ui.userfeedback.choice(self.header, c, canceltext="Close", choiceimages=cimgs)
            if ic is not None:
                idx2act[ic](ais)
                ais.save()
            else:
                break

    # noinspection PyUnusedLocal
    def getcustomactions(self, ais: anise.framework.projects.IntermediateStructure) -> List['CustomAction']:
        """
        Returns a list of custom actions, controlling which additional actions are offered in the overview.

        :param ais: A project structure.
        """
        return []

    # noinspection PyUnusedLocal
    def getentries(self, ais: anise.framework.projects.IntermediateStructure
                   ) -> List[anise.framework.projects.IntermediateStructure.Node]:
        """
        Returns a list of nodes to show in the overview.
        |dooverride_optional|

        :param ais: A project structure.
        """
        return []

    # noinspection PyUnusedLocal
    def getentrylabel(self, ais: anise.framework.projects.IntermediateStructure,
                      node: anise.framework.projects.IntermediateStructure.Node) -> str:
        """
        Returns a label text for a node.
        |dooverride_optional|

        :param ais: A project structure.
        :param node: A project structure node.
        """
        return str(node)

    def manageentry(self, ais: anise.framework.projects.IntermediateStructure,
                    node: anise.framework.projects.IntermediateStructure.Node) -> None:
        """
        Called for managing one of the entries.

        The default implementation opens the standard node editor (as the raw mode editor would provide it).

        :param ais: A project structure.
        :param node: A project structure node.
        """
        epa = universe.projectdescriptioneditor.Internals.EditProjectdescApplication()
        epa.init(ais)
        epa.changeentry(node, remove_is_unsafe=False)
