# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

import anise.framework.exceptions
import anise.framework.features
import anise.framework.projects

from anise.framework.projects import universe


pde_distributables = anise.framework.features.loadfeature("projectdescriptioneditor.distributables").featuremodule
projectdescriptioneditor = anise.framework.features.loadfeature("projectdescriptioneditor").featuremodule


class AbstractPackage:

    def __init__(self):
        self.label = None
        self.taskref = None

    def gettaskref(self):
        raise anise.framework.exceptions.NotImplementedError()

    def ask(self):
        raise anise.framework.exceptions.NotImplementedError()


@anise.framework.features.hookhandler(pde_distributables.HOOK_GET_DISTRIBUTABLES_PROVIDERS)
class TarPackage(AbstractPackage):

    def __init__(self):
        AbstractPackage.__init__(self)
        self.label = "source tar package"
        self.taskref = "packages.tarpackage"

    def ask(self):
        return anise.framework.projects.IntermediateStructure.Value("dict")


@anise.framework.features.hookhandler(pde_distributables.HOOK_GET_DISTRIBUTABLES_PROVIDERS)
class DebPackage(AbstractPackage):

    def __init__(self):
        AbstractPackage.__init__(self)
        self.label = "Debian package"
        self.taskref = "packages.debian.debpackage"

    def ask(self):
        return anise.framework.projects.IntermediateStructure.Value("dict")


@anise.framework.features.hookhandler(pde_distributables.HOOK_GET_DISTRIBUTABLES_PROVIDERS)
class PythonWheelPackage(AbstractPackage):

    def __init__(self):
        AbstractPackage.__init__(self)
        self.label = "Python wheel package"
        self.taskref = "packages.python.wheelpackage"

    def ask(self):
        return anise.framework.projects.IntermediateStructure.Value("dict")


@anise.framework.features.hookhandler(pde_distributables.HOOK_GET_DISTRIBUTABLES_PROVIDERS)
class Win32Package(AbstractPackage):

    def __init__(self):
        AbstractPackage.__init__(self)
        self.label = "Win32 installer package"
        self.taskref = "packages.win32.win32exepackage"

    def ask(self):
        projecticon = universe.ui.userfeedback.input(
            "Please enter a path (relative to the project root directory) to a Windows icon file (optional).",
            "_meta/{0}.ico".format(universe.name))
        if projecticon is None:
            return None
        return anise.framework.projects.IntermediateStructure.Value(
            "dict", projecticon=anise.framework.projects.IntermediateStructure.Value.fromstring(projecticon))


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(
    projectdescriptioneditor.HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS)
class Win32PackageAddStartMenuEntryCustomAction(projectdescriptioneditor.CustomAction):

    def __init__(self):
        super().__init__("Add start menu entry", icon="userfeedback/add.png")

    def visible(self, ais, node):
        if node.isvalue and node.vtype == "files.TaskExecution":
            t = node.getarg("task")
            return t and t.oref == "packages.win32.win32exepackage"

    def execute(self, ais, node):
        command = universe.ui.userfeedback.input(
            "Please specify the executable to be executed (as absolute path on the target "
            "machine; $$: installation directory).", fr"$$\{universe.name}.exe")
        if command is None:
            return None
        params = universe.ui.userfeedback.input("Please specify the commandline parameters.", "")
        if params is None:
            return None
        title = universe.ui.userfeedback.input("Please specify a label for the start menu entry.", universe.name)
        if title is None:
            return None
        icon = universe.ui.userfeedback.input(
            "Please enter a path (relative to the project root directory) to a Windows icon file (optional).",
            "_meta/{0}.ico".format(universe.name))
        if icon is None:
            return None
        if "menuentries" in node.args():
            ell = node.getarg("menuentries")
        else:
            ell = anise.framework.projects.IntermediateStructure.Value("list")
            node.setarg("menuentries", ell)
        oprms = {}
        if len(icon.strip()) > 0:
            oprms["icon"] = anise.framework.projects.IntermediateStructure.Value.fromstring(icon)
        if len(params.strip()) > 0:
            oprms["params"] = anise.framework.projects.IntermediateStructure.Value.fromstring(params)
        vapplink = anise.framework.projects.IntermediateStructure.Value(
            "packages.win32.MenuEntry",
            command=anise.framework.projects.IntermediateStructure.Value.fromstring(command),
            title=anise.framework.projects.IntermediateStructure.Value.fromstring(title), **oprms)
        ell.appendarg(vapplink)


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(
    projectdescriptioneditor.HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS)
class DebPackageAddStartMenuEntry(projectdescriptioneditor.CustomAction):

    def __init__(self):
        super().__init__("Add start menu entry", icon="userfeedback/add.png")

    def visible(self, ais, node):
        if node.isvalue and node.vtype == "files.TaskExecution":
            t = node.getarg("task")
            return t and t.oref == "packages.debian.debpackage"

    def execute(self, ais, node):
        command = universe.ui.userfeedback.input("Please specify the command to be associated with the entry.",
                                                 universe.name)
        if command is None:
            return None
        title = universe.ui.userfeedback.input("Please specify a label for the start menu entry.", universe.name)
        if title is None:
            return None
        icon = universe.ui.userfeedback.input(
            "Please enter a path (relative to the project root directory) to the icon (optional).",
            "_meta/{0}.64.png".format(universe.name))
        if icon is None:
            return None
        categories = sorted([x for x in dir(universe.packages.debian.Category) if not x.startswith("_")])
        scategories = [getattr(universe.packages.debian.Category, x)[0] for x in categories]
        icategory = universe.ui.userfeedback.choice("Please choose a program category.", choices=scategories)
        if icategory is None:
            return None
        isgui = universe.ui.userfeedback.messageyesno(
            "Does the command start a graphical tool (which does not need a terminal)?")
        if "menuentries" in node.args():
            ell = node.getarg("menuentries")
        else:
            ell = anise.framework.projects.IntermediateStructure.Value("list")
            node.setarg("menuentries", ell)
        value = anise.framework.projects.IntermediateStructure.Value
        oprms = {'name': value.fromstring(universe.name + str(len(ell.args()))), 'command': value.fromstring(command),
                 'title': value.fromstring(title), 'gui': value.frombool(isgui),
                 'category': value("objectref", oref="packages.debian.Category." + categories[icategory])}
        if len(icon.strip()) > 0:
            oprms["icon"] = anise.framework.projects.IntermediateStructure.Value.fromstring(icon)
        vapplink = anise.framework.projects.IntermediateStructure.Value("packages.debian.MenuEntry", **oprms)
        ell.appendarg(vapplink)


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(
    projectdescriptioneditor.HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS)
class DebPackageAddExecLinkCustomAction(projectdescriptioneditor.CustomAction):

    def __init__(self):
        super().__init__("Add executable link for easier access", icon="userfeedback/add.png")

    def visible(self, ais, node):
        if node.isvalue and node.vtype == "files.TaskExecution":
            t = node.getarg("task")
            return t and t.oref == "packages.debian.debpackage"

    def execute(self, ais, node):
        linkname = universe.ui.userfeedback.input("Please specify a name for the link (typically an easy one).",
                                                  universe.name)
        if linkname is None:
            return None
        linktarget = universe.ui.userfeedback.input(
            "Please specify the destination path for the link (on the target machine).", "/usr/bin/foo.bin")
        if linktarget is None:
            return None
        if "executablelinks" in node.args():
            ell = node.getarg("executablelinks")
        else:
            ell = anise.framework.projects.IntermediateStructure.Value("dict")
            node.setarg("executablelinks", ell)
        ell.setarg(linkname, anise.framework.projects.IntermediateStructure.Value.fromstring(linktarget))


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(
    projectdescriptioneditor.HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS)
class PythonWheelPackageAddExecLinkCustomAction(projectdescriptioneditor.CustomAction):

    def __init__(self):
        projectdescriptioneditor.CustomAction.__init__(
            self, "Add application link for starting a Python method from command-line", icon="userfeedback/add.png")

    def visible(self, ais, node):
        if node.isvalue and node.vtype == "files.TaskExecution":
            t = node.getarg("task")
            return t and t.oref == "packages.python.wheelpackage"

    def execute(self, ais, node):
        linkname = universe.ui.userfeedback.input("Please specify a name for the link (typically an easy one).",
                                                  universe.name)
        if linkname is None:
            return None
        module = universe.ui.userfeedback.input(
            "Please specify the Python module name that contains the runnable function.",
            universe.name + ".foo.engine")
        if module is None:
            return None
        fctname = universe.ui.userfeedback.input("Please specify the function that should be called by that link.",
                                                 "main")
        if fctname is None:
            return None
        isgui = universe.ui.userfeedback.messageyesno(
            "Does this method trigger a gui application (which does not need a terminal)?")
        if "applicationlinks" in node.args():
            ell = node.getarg("applicationlinks")
        else:
            ell = anise.framework.projects.IntermediateStructure.Value("list")
            node.setarg("applicationlinks", ell)
        vapplink = anise.framework.projects.IntermediateStructure.Value(
            "packages.python.WheelApplicationLink",
            linkname=anise.framework.projects.IntermediateStructure.Value.fromstring(linkname),
            module=anise.framework.projects.IntermediateStructure.Value.fromstring(module),
            method=anise.framework.projects.IntermediateStructure.Value.fromstring(fctname),
            gui=anise.framework.projects.IntermediateStructure.Value("bool", v=isgui))
        ell.appendarg(vapplink)
