# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for a particular :samp:`release` task. It collects actions to execute for releasing, e.g. generating and
deploying a project homepage, by triggering a hook. The :samp:`release` task is provided, which bundles all those
actions.
"""

from typing import *

import anise.framework.features
import anise.framework.report

from anise.framework.projects import universe


#: Hook triggered for collecting all sub-tasks for :samp:`release`.
HOOK_RELEASE = anise.framework.features.Hook(
    interesting=0, doc="This hook is called for collecting all sub-tasks for 'release'.")


class Internals:

    class Releasetasks:
        """
        Storage of all actions to execute for releasing.
        """

        def add(self, task: Callable, *, provides: Optional[Union[str, List[str]]] = None,
                requires: Optional[Union[str, List[str]]] = None,
                prepares: Optional[Union[str, List[str]]] = None, **params):
            """
            Specifies a new task and adds it into the releasing chain.

            :param task: The task implementation to add.
            :param provides: A list of symbol names this hook participates to provide (dependency provider).
            :param requires: A list of symbol names this hook requires to be executed before.
            :param prepares: A list of symbol names this hook must prepend.
            :param params: Additional parameters for calling that task.
            """
            universe.addhook(HOOK_RELEASE, task, params, provides, requires, prepares)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject1():
        universe.releasing.tasks = Internals.Releasetasks()

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _initproject2():
        if len(universe.gethooks(HOOK_RELEASE)) > 0:
            universe.addtask(Internals.release, "releasing.")

    @staticmethod
    def release() -> None:
        """
        Releases the project. Mostly this is executing all registered hookhandlers for :py:data:`HOOK_RELEASE`.
        """
        with universe.executionscope.newsubscope("Releasing") as scope:
            hooks = universe.gethooks(HOOK_RELEASE)
            for i, fct in enumerate(hooks):
                scope.progress = anise.framework.report.ExecutionScope.ProgressDeterminate((float(i)/len(hooks)))
                fct()
