# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for signing binaries with cryptographic certificates.
"""

from typing import *
import os

import anise.framework.exceptions
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


class Internals:

    @staticmethod
    def win32_osslsigncode(*, binpath: str, pempath: str, outpath: str, projectname: str,
                           projecthomepage: Optional[str], osslsigncodepath: str, password: str) -> None:
        """
        Signs a win32 binary with osslsigncode.

        :param binpath: Path to the binary to sign.
        :param pempath: Path to the pem certificate file to use.
        :param outpath: Destination path for the signed binary.
        :param projectname: The project name.
        :param projecthomepage: The project homepage url.
        :param osslsigncodepath: Path to the osslsigncode tool.
        :param password: Certificate password.
        """
        if anise.utils.basic.call(osslsigncodepath + " -v", shell=True)[0] not in [0, 255]:
            raise anise.framework.exceptions.RequirementsMissingError("'osslsigncode' seems to be unavailable")
        callparms = [osslsigncodepath, "sign", "-spc", pempath, "-key", pempath, "-n", projectname, "-in", binpath,
                     "-out", outpath, "-pass", password]
        if projecthomepage:
            callparms += ["-i", projecthomepage]
        r, o = anise.utils.basic.call(*callparms)
        if r != 0:
            raise anise.framework.exceptions.ProcessExecutionFailedError("unable to run osslsigncode", o)
        anise.utils.logging.logdebug(o)

    @staticmethod
    def p12cert2pem(*, opensslpath: str, p12path: str, pemoutpath: str, password: str) -> None:
        """
        Converts a .p12 certificate to a :file:`.pem` file.

        :param opensslpath: Path to the OpenSSL binary.
        :param p12path: Path to the p12 certificate file.
        :param pemoutpath: Destination path for the pem file.
        :param password: Certificate password of the p12 file.
        """
        if anise.utils.basic.call(opensslpath + " version", shell=True)[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'openssl' seems to be unavailable")
        r, o = anise.utils.basic.call(opensslpath, "pkcs12", "-in", p12path, "-out", pemoutpath, "-passin",
                                      "pass:"+password, "-passout", "pass:"+password)
        if r != 0:
            raise anise.framework.exceptions.ProcessExecutionFailedError("unable to run openssl", o)
        anise.utils.logging.logdebug(o)


class OsslsigncodeConfiguration:
    """
    Signing configuration for signing a win32 binary with the 'osslsigncode' tool.
    """

    def __init__(self, p12path: str, intermediatepath: Optional[str] = None, osslsigncodepath: str = "osslsigncode",
                 opensslpath: str = "openssl", password: str = ""):
        """
        :param p12path: Path to the .p12 file.
        :param intermediatepath: Path to intermediate certificates.
        :param osslsigncodepath: Path to the osslsigncode tool.
        :param opensslpath: Path to the openssl tool.
        :param password: The certificate passphrase.
        """
        self.osslsigncodepath = osslsigncodepath
        self.opensslpath = opensslpath
        self.p12path = p12path
        self.intermediatepath = intermediatepath
        self.password = password
        

def signwin32_osslsigncode(signconfig: OsslsigncodeConfiguration, source: anise.framework.files.Filestructure,
                           required: bool = True) -> anise.framework.files.Filestructure:
    """
    Signs a win32 binary with a cryptographic code-signing certificate.

    :param signconfig: A configuration object.
    :param source: The source file structure.
    :param required: If a failure is critical or can be ignored.
    """
    with universe.executionscope.newsubscope("Creating a signed win32 binary"):
        targetdir = anise.framework.files.TempDir()
        outpath = targetdir.path + "/" + os.path.basename(source.path)
        projectname = universe.name
        pempath = targetdir.path + "/crt.der"
        try:
            Internals.p12cert2pem(
                opensslpath=signconfig.opensslpath, p12path=signconfig.p12path, pemoutpath=pempath,
                password=signconfig.password)
            if signconfig.intermediatepath:
                with open(signconfig.intermediatepath, "r") as f:
                    intermediatepem = f.read()
                with open(pempath, "a") as f:
                    f.write("\n" + intermediatepem)
            Internals.win32_osslsigncode(
                binpath=source.path, pempath=pempath, outpath=outpath, projectname=projectname,
                projecthomepage=universe.base.websiteurl, osslsigncodepath=signconfig.osslsigncodepath,
                password=signconfig.password)
            return anise.framework.files.Filestructure(outpath)
        except Exception as e:
            if required:
                raise
            else:
                anise.utils.logging.logwarning("Unable to sign win32 binary (leaving it unsigned): " + str(e))
                return source
