# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for terminal-based user interfaces.
"""

import math
import os
import subprocess
import sys

import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

ui = anise.framework.features.loadfeature("ui").featuremodule


class Internals:

    # noinspection PyUnusedLocal
    class UserFeedback(ui.Internals.UserFeedback):
        """
        Uses the standard input and standard output channel of the terminal for user interaction.
        """

        def __init__(self):
            super().__init__()
            self.termhelper = anise.framework.features.loadfeature("ui.terminal.helpers").featuremodule
            # crazy workaround for a crazy terminal problem i don't understand :)
            if sys.stderr:
                stderr = getattr(sys.stderr, "originalstream", sys.stderr)
                if stderr:
                    stderr.write("\n\n\n\n\n\n\033[30;104m\n\n\n\n\033[30;107m\n\n\n\n\033[33;107m""")
                    stderr.write("\033[0m\n\n\n\n")

        def message(self, msg, btns, image):
            if btns == ["OK"]:
                self.termhelper.prompt(msg, "Please press Enter:")
                return 0
            else:
                return self.choice(msg, btns, False, image)

        def input(self, msg, default, image):
            if default and len(default) > 0:
                sdefault = "\nDefault suggestion (you have to explicitely type it!): '" + default + "'"
            else:
                sdefault = ""
            ans = self.termhelper.prompt(msg, sdefault + "\nPlease press Enter afterwards (enter '-c-' for cancel):")
            if ans == "-c-":
                return None
            return ans

        def multilineinput(self, msg, default, image):
            emsg = """
            After you press Enter, a text editor will start. When you are done,
            please just 'Save' the file and close the editor afterwards.
            """
            self.termhelper.prompt(msg, emsg)
            if "EDITOR" in os.environ:
                editor = os.environ["EDITOR"]
            else:
                editor = "nano"
            with anise.framework.files.TempDir() as tempdir:
                tmpfile = tempdir.path + "/aniseui.val"
                anise.utils.basic.writetofile(tmpfile, default)
                subprocess.call([editor, tmpfile])
                nv = anise.utils.basic.readfromfile(tmpfile).rstrip("\r\n")
            return nv

        def treechoice(self, msg, choicetree, canceltext, actions, image):
            choices = []
            stack = []
            idtoresult = []
            for x in choicetree.root.childs:
                stack.append((x, 0))
            stack.reverse()
            while len(stack) > 0:
                node, indent = stack.pop()
                key = (indent * "    ") + node.key
                if len(key) > 65:
                    key = key[:61] + " ..."
                if "\n" in key:
                    key = key[:key.find("\n")] + " ..."
                choices.append(key)
                idtoresult.append(node)
                rnodechilds = list(node.childs)
                rnodechilds.reverse()
                for c in rnodechilds:
                    stack.append((c, indent+1))
            for sbtn in actions:
                choices.append(str(sbtn))
                idtoresult.append(sbtn)
            ic = self.choice(msg, choices, canceltext, image)
            if ic is None:
                return None
            else:
                return idtoresult[ic]

        def choice(self, msg, choices, canceltext, image, choiceimages=None):
            choicenummaxlen = int(math.floor(math.log10(len(choices)))) + 1
            schoices = "\n"
            for i, c in enumerate(choices):
                sc = str(c)
                num = (" "*choicenummaxlen + str(i+1))[-choicenummaxlen:]
                schoices += f"    {num}. {sc}\n"
            if canceltext:
                sc = canceltext
                num = (" "*choicenummaxlen + "c")[-choicenummaxlen:]
                schoices += f"\n    {num}. {sc}\n"
            choice = None
            while choice is None:
                choice = self.termhelper.prompt(msg, f"{schoices}\nPlease make a choice and press Enter:")
                if canceltext and choice.lower() == "c":
                    choice = None
                    break
                try:
                    choice = int(choice) - 1
                    if choice < 0 or choice >= len(choices):
                        choice = None
                except Exception:
                    choice = None
            return choice

    @staticmethod
    @anise.framework.features.hookhandler(ui.HOOK_UIMODE_SELECTED, provides=["PRE"])
    def uimodeselected():
        userfeedback = Internals.UserFeedback()
        universe.addhook(ui.HOOK_GET_USERFEEDBACK_IMPLEMENTATION, lambda: userfeedback, provides=[ui.Mode.TERMINAL])
