# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for Webexec, which is the default web interface for task executions.
"""

import ctypes
import threading

import anise.data.Yaji.yaji as yaji
import anise.features.ui_.web.helpers
import anise.framework.features
import anise.framework.globalvars
import anise.framework.report

from anise.framework.projects import universe

ui = anise.framework.features.loadfeature("ui").featuremodule


class WebexecApplication(anise.features.ui_.web.helpers.AniseWebApplication):

    class AbortedInterrupt(KeyboardInterrupt):

        def __init__(self):
            super().__init__()

        def __str__(self):
            return "Aborted by user."

    def __init__(self, **b):
        super().__init__(head1=f"Anise Task Execution: {universe.name}",
                         show_browser_closed_notification=True, skip_shutdown_dialog=True, **b)
        self.startmessageonlyterminal = True
        self.mainthread = threading.current_thread()
        self.mainexecscope = anise.framework.report.topscope()
        self.try_addclientresources_from_default_location(__file__)
        self.universe = None
        self._shutdownbyuser = False
        self.bodyleftviewactionlabel = "Back to user interaction"
        self.bodyrightviewactionlabel = "Execution pager"
        self.dsshowdebug = self.create_datastore()
        self.dsshowdebug.add_onchanged_handler(self._update_menu)
        self.dsshowdebug.setvalue(bool(anise.framework.globalvars.logdebug))
        self.setbodyright(yaji.View('anise.ExecPager', hstretch=1, vstretch=1, name='pager',
                                    debugVisible=self.binddata(self.dsshowdebug)))
        self.head2 = yaji.View('anise.WebexecHeader', hstretch=1)

    def _update_menu(self, *_):
        actions = []
        if self.dsshowdebug.getvalue():
            actions.append(yaji.Action('Hide details', yaji.BackendFunction('showlogdebug', {'value': False})))
        else:
            actions.append(yaji.Action('Show details', yaji.BackendFunction('showlogdebug', {'value': True})))
        actions.append(yaji.Action('Abort execution', yaji.BrowserFunction('()=>yaji.stop()')))
        self.setactions(actions)

    def oninitialize(self):
        super().oninitialize()
        self.universe = universe

    def stop(self):
        stopimplicit = self.stop_implicitly_when_browser_closed
        if self._shutdownbyuser or stopimplicit:
            exc = ctypes.py_object(WebexecApplication.AbortedInterrupt)
            ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self.mainthread.ident), exc)
            self.mainthread.join()
            if stopimplicit:  # just for the case anyone unset that meanwhile (now there is no way back)
                self.stop_implicitly_when_browser_closed = True
        super().stop()

    @yaji.RequestHandler.for_urls(auto=True)
    def _do_showlogdebug(self, value: bool):
        self.dsshowdebug.setvalue(value)

    @yaji.RequestHandler.for_urls(auto=True)
    def _do__yj_application_stop(self):
        # noinspection PyProtectedMember
        self._shutdownbyuser = True
        return super()._do__yj_application_stop()


@anise.framework.features.hookhandler(ui.HOOK_UIMODE_SELECTED)
def initproject():
    if universe.ui.mode == ui.Mode.WEB:
        ui.web.webexec.instance = WebexecApplication(parentid=getattr(ui.web.webexec, "parentid", None),
                                                     returntoparent=getattr(ui.web.webexec, "returntoparent", True))
        class SetImplicitStopAllowed(ui.Internals.SetImplicitStopAllowed):
            def __init__(self, val):
                self.val = val
            def __enter__(self):
                self.o = ui.web.webexec.instance.do_stop_implicitly_when_browser_closed(self.val)
                return self.o.__enter__()
            def __exit__(self, exc_type, exc_val, exc_tb):
                return self.o.__exit__(exc_type, exc_val, exc_tb)
        universe.addhook(ui.HOOK_GET_ALLOWIMPLICITCLOSE_IMPLEMENTATION, SetImplicitStopAllowed, prepares=["ui"])
        ui.web.webexec.instance.start()
