# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Features for interaction with the PiGet version control system.
"""

import os
import shutil

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

versioncontrol = anise.framework.features.loadfeature("versioncontrol").featuremodule


# noinspection PyUnusedLocal
class PigetVersionControlSystem(versioncontrol.Internals.VersionControlSystem):
    """
    PiGet version control system adapter.
    """

    def __init__(self):
        versioncontrol.Internals.VersionControlSystem.__init__(self)
        self.label = "Git"

    def getfullversion(self):
        _version = getattr(universe.base, "version", None)
        return ((_version + ".") if _version is not None else "") + str(self.getheadrevision())

    def createrawpackage(self):
        import piget
        res = []
        with universe.executionscope.newsubscope("Fetching the sources from PiGet working copy"):
            tmp = anise.framework.files.TempDir("rawpackage")
            def ohfile(fullsx, fulldx, relx, stype, dtype):
                shutil.copyfile(fullsx, fulldx)
            def ohlink(fullsx, fulldx, relx, stype, dtype):
                os.symlink(os.readlink(fullsx), fulldx)
            piget.enumeratepackage(universe.specialpaths.projectdir, tmp.path,
                                   skippaths=['/.piget', '/README'], onhandlefile=ohfile, onhandlelink=ohlink)
            return tmp

    def syncversioncontrolsystem(self, commitmessage="", forceskipped=True, forceunchanged=False):
        if forceskipped or (not getattr(universe, "_vcssynced", False)):
            anise.utils.logging.loginfo("Synchronizing the working copy with the piget repository.", printcaller=False)
            with anise.utils.basic.ChDir(universe.specialpaths.projectdir):
                import piget
                piget.piget_commit(universe.specialpaths.projectdir)
            universe._vcssynced = True

    def getheadrevision(self):
        import piget
        return int(piget.getcontrolvalue("version", universe.specialpaths.projectdir))

    def fetchrawcommitmessage(self, revision):
        return f"{revision}"

    def getcommitmessagefromrawcommitmessage(self, s):
        return versioncontrol.CommitMessage("", int(s), None, None)

    def branchfromhere(self, branchname=None):
        raise NotImplementedError()
