# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Features for interaction with the Subversion version control system.
"""

from typing import *
import datetime
import os
import re
import shutil

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

pdeversioncontrol = anise.framework.features.loadfeature("projectdescriptioneditor.versioncontrol").featuremodule
ui = anise.framework.features.loadfeature("ui").featuremodule
versioncontrol = anise.framework.features.loadfeature("versioncontrol").featuremodule


class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(pdeversioncontrol.HOOK_GET_VERSIONCONTROLSYSTEM_IMPLEMENTATIONS)
    def _register_vcs():
        return SubversionVersionControlSystem, "versioncontrol.svn", "Subversion"

    class Version:
        """
        Version number of an Anise Subversion project.
        """
        def __init__(self, build: int, tag: str):
            self.build = build
            self.tag = tag

        def isvalid(self) -> bool:
            return self.tag != "unknown"

        def __str__(self):
            return str(self.build) + (f"+{self.tag}" if self.tag else "")

    @staticmethod
    def getsvnrevision(dirpath: str) -> 'Internals.Version':
        """
        Determines the revision number of a Subversion working copy.

        :param dirpath: The directory path.
        """
        try:
            with anise.utils.basic.ChDir(dirpath):
                o, s = anise.utils.basic.call("svnversion", raise_on_errors="unable to call svn")
            m = re.search(r"^([0-9]+)(:[0-9]+)?([A-Z]+)?", s)
            build = int(m.group(1))
            changed = (m.group(2) is not None) or (m.group(3) is not None)
            tag = None
            if changed:
                tag = "alike"
        except anise.framework.exceptions.ProcessExecutionFailedError:
            build = 0
            tag = "unknown"
        return Internals.Version(build, tag)

    @staticmethod
    def listsvnrepo(dirpath: str) -> List[str]:
        """
        Determines the file tree of an Subversion repository.

        :param dirpath: The directory path.
        """
        (r, o) = anise.utils.basic.call("svn", "list", "-R", dirpath, "-r", "HEAD",
                                        raise_on_errors="unable to call svn")
        res = [x for x in [a.strip() for a in o.split("\n")] if x != ""]
        return res


# noinspection PyUnusedLocal
class SubversionVersionControlSystem(versioncontrol.Internals.VersionControlSystem):
    """
    Subversion version control system adapter.
    """

    def __init__(self):
        super().__init__()

    def getfullversion(self):
        _version = getattr(universe.base, "version", None)
        _revision = str(Internals.getsvnrevision(universe.specialpaths.projectdir))
        return (f"{_version}." if _version is not None else "") + _revision

    def createrawpackage(self):
        with universe.executionscope.newsubscope("Fetching the sources from subversion working copy"):
            tmp = anise.framework.files.TempDir("rawpackage")
            for e in Internals.listsvnrepo(universe.specialpaths.projectdir):
                full = f"{universe.specialpaths.projectdir}/{e}"
                tgtfull = f"{tmp.path}/{e}"
                if os.path.islink(full):
                    os.symlink(os.readlink(full), tgtfull)
                elif os.path.isdir(full):
                    os.mkdir(tgtfull)
                elif os.path.isfile(full):
                    shutil.copy2(full, tgtfull)
                else:
                    raise anise.framework.exceptions.BadFormatError(f"unknown thing: {full}")
            return tmp

    def syncversioncontrolsystem(self, commitmessage="", forceskipped=True, forceunchanged=False):
        if forceskipped or (not getattr(universe, "_vcssynced", False)):
            anise.utils.logging.loginfo("Synchronizing the working copy with the svn repository.", printcaller=False)
            with anise.utils.basic.ChDir(universe.specialpaths.projectdir):
                anise.utils.basic.call("svn", "update", raise_on_errors="errors in svn")
                if forceunchanged:
                    anise.utils.basic.call("svn", "propset", "anise_tmp", str(datetime.datetime.now().timestamp()), ".")
                anise.utils.basic.call("svn", "commit", "-m", commitmessage, raise_on_errors="errors in svn")
                if forceunchanged:
                    anise.utils.basic.call("svn", "propdel", "anise_tmp", ".")
                anise.utils.basic.call("svn", "update", raise_on_errors="errors in svn")
            universe._vcssynced = True

    def findvcs(self):
        return os.path.isdir(f"{universe.specialpaths.projectdir}/.svn")

    def bindvcs(self, l):
        url = universe.ui.userfeedback.input("Please enter the svn repository url.", f"https://foo/svn/{universe.name}")
        if url is not None:
            with anise.framework.files.TempDir() as td:
                # no error handling here
                anise.utils.basic.call("svn", "import", "-m", "anise initial import", td.path, url)
            with anise.utils.basic.ChDir(universe.specialpaths.projectdir):
                if os.path.exists(universe.specialpaths.projectdir + "/.svn"):
                    shutil.rmtree(universe.specialpaths.projectdir + "/.svn")
                anise.utils.basic.call("svn", "co", url, ".", raise_on_errors="errors in svn")
                anise.utils.basic.call("svn add *", raise_on_errors="errors in svn", shell=True)
                anise.utils.basic.call("svn", "commit", "-m", "anise initial commit", raise_on_errors="errors in svn")
                anise.utils.basic.call("svn", "update", raise_on_errors="errors in svn")
            return True

    def getheadrevision(self):
        return Internals.getsvnrevision(universe.specialpaths.projectdir).build

    def fetchrawcommitmessage(self, revision):
        with anise.utils.basic.ChDir(universe.specialpaths.projectdir):
            r, res = anise.utils.basic.call("svn", "log", ".@"+str(revision), "-l1", raise_on_errors="errors in svn")
            return res

    def getcommitmessagefromrawcommitmessage(self, s):
        rcommitmsg = [x.rstrip() for x in s.split("\n")]
        while rcommitmsg[-1] == "":
            rcommitmsg = rcommitmsg[:-1]
        resmessage = "\n".join(rcommitmsg[2:-1])
        mdparts = [x for x in [x.strip() for x in rcommitmsg[1].split(" | ")] if x]
        resrevision = re.search(r'\d+', mdparts[0]).group(0)
        rescommittime = datetime.datetime.fromtimestamp(
            datetime.datetime.strptime(mdparts[2][0:25], "%Y-%m-%d %H:%M:%S %z").timestamp())
        resusername = mdparts[1]
        return versioncontrol.CommitMessage(resmessage, resrevision, rescommittime, resusername)
