# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
The Anise plug-in system.

This package does not include the Anise high-level features itself. You would find them in :py:mod:`anise.features`.
This package is the part of the Anise infrastructure that loads and manages those features.

For each Anise run, it automatically loads the built-in :py:mod:`anise.features` as well as custom features from
:file:`~/.anise/features` and :file:`/var/lib/anise/features`. Due to their locations, the latter ones are useful for
per-machine and per-user customization.

In an arbitrary Anise execution context, you may call :py:func:`anise.framework.features.loadfeature` in order to access
functionality from other features, like here:

.. code-block:: python

  from anise.framework import features
  _somefeature = features.loadfeature("somefeature").featuremodule
  _somefeature.some_method()
  x = _somefeature.some_variable
"""

from typing import *
import importlib
import importlib.machinery
import importlib.util
import os

import anise.framework.exceptions
import anise.utils


class Internals:
    """
    Internal stuff.
    """

    @staticmethod
    def inject_defaults_to_universe(u: 'projects.Universe') -> None:
        """
        Adds feature data structures to the universe object.
        """
        for feature in _features:
            fname = feature.name if ("." not in feature.name) else feature.name[feature.name.rindex(".")+1:]
            setattr(anise.utils.data.createobjectstructure(u, feature.name, PseudoFeature)[1], fname,
                    feature.featuremodule)

    @staticmethod
    def inject_hooks_to_universe(u: 'projects.Universe') -> None:
        """
        Adds hooks to the universe object, which are loaded in feature loading.
        """
        for feature in _features:
            # noinspection PyProtectedMember
            for h in feature._hookhandlers:
                u.addhook(h.hookobj, h.function, None, h.provides, h.requires, h.prepares)

    @staticmethod
    def loadfeatures(featurespath: str) -> None:
        """
        Loads features from a path.
        """
        ls = [(featurespath, "")]
        ll = []
        while len(ls) > 0:
            xpath, xname = ls.pop()
            if os.path.isdir(xpath):
                for f in sorted(os.listdir(xpath)):
                    if f.startswith("__"):
                        continue
                    nname = xname
                    if nname != "":
                        nname += "."
                    nname += f
                    if nname.endswith(".py"):
                        nname = nname[:-3]
                    elif nname.endswith("_"):
                        nname = nname[:-1]
                    ls.append((f"{xpath}/{f}", nname))
            elif os.path.isfile(xpath):
                _featuresources.append((xpath, xname))
                ll.append(xname)
        for xxname in ll:
            loadfeature(xxname)

    @staticmethod
    def resetfeatures() -> None:
        """
        Resets the features.
        """
        global _features
        global _featuresources
        anise.utils.logging.logdebug("Resetting features.")
        _features = []
        _featuresources = []


class HookKind:
    """
    Enumeration of hook kinds.
    """

    #: A function kind hook.
    FUNCTION = "FUNCTION"

    #: A class kind hook.
    CLASS = "CLASS"


class Hook:
    """
    A hook. See documentation for details.
    """

    def __init__(self, hookkind: str = HookKind.FUNCTION, *, hide: bool = False, interesting: int = 5,
                 doc: Optional[str] = None, wizard_functionparams: Optional[List[str]] = None,
                 wizard_functionadditionallog: Optional[List[str]] = None, wizard_functionbody: Optional[str] = None,
                 wizard_classbase: Optional[str] = None, wizard_classbody: Optional[str] = None,
                 wizard_loadfeatures: Optional[List[str]] = None):
        """
        :param hookkind: A hook kind (one of :py:class:`HookKind`).
        :param hide: If this hook is to be hidden in hook choosers.
        :param interesting: A number that controls how interesting this hook is in a chooser.
        :param doc: Hook documentation text.
        :param wizard_functionparams: Info for hook handler wizards. List of function parameters.
        :param wizard_functionadditionallog: Info for hook handler wizards. List of additional variable to log out in a
                                             function.
        :param wizard_functionbody: Info for hook handler wizards. Additional function body part.
        :param wizard_classbase: Info for hook handler wizards. Base class.
        :param wizard_classbody: Info for hook handler wizards. Additional class body part.
        :param wizard_loadfeatures: Info for hook handler wizards. Additional features to load.
        """
        self.hookkind = hookkind
        self.hide = hide
        self.interesting = interesting
        self.__doc__ = doc
        self.wizard_functionparams = wizard_functionparams or []
        self.wizard_functionadditionallog = wizard_functionadditionallog or []
        self.wizard_functionbody = wizard_functionbody
        self.wizard_classbase = wizard_classbase
        self.wizard_classbody = wizard_classbody
        self.wizard_loadfeatures = wizard_loadfeatures or []


class HookHandler:
    """
    A hook handler bound to some hook. See documentation for details.
    """

    def __init__(self, hookobj: Hook, function: Callable,
                 provides: Optional[Union[str, List[str]]], requires: Optional[Union[str, List[str]]],
                 prepares: Optional[Union[str, List[str]]]):
        """
        :param hookobj: The hook.
        :param function: The handler function/class.
        :param provides: A list of symbol names this hook participates to provide (dependency provider).
        :param requires: A list of symbol names this hook requires to be executed before.
        :param prepares: A list of symbol names this hook must prepend.
        """
        self.__hookobj = hookobj
        self.__function = function
        self.__provides = list(provides.split(" ") if isinstance(provides, str) else (provides or []))
        self.__requires = list(requires.split(" ") if isinstance(requires, str) else (requires or []))
        self.__prepares = list(prepares.split(" ") if isinstance(prepares, str) else (prepares or []))

    @property
    def hookobj(self) -> Hook:
        """
        The hook.
        """
        return self.__hookobj

    @property
    def function(self) -> Callable:
        """
        The handler function/class.
        """
        return self.__function

    @property
    def provides(self) -> List[str]:
        """
        The symbol names this hook participates to provide (dependency provider).
        """
        return self.__provides

    @property
    def requires(self) -> List[str]:
        """
        The symbol names this hook requires to be executed before.
        """
        return self.__requires

    @property
    def prepares(self) -> List[str]:
        """
        The symbol names this hook must prepend.
        """
        return self.__prepares


class PseudoFeature:
    """
    This is an auxiliary class, which fixes the gap that appears, e.g. if a feature :samp:`foo.bar` exists, but
    no :samp:`foo` one.
    """
    pass


class Feature:
    """
    A feature is some encapsuled part of anise high-level functionality. This class represents a loaded feature
    at runtime. The framework as well as other features access this object for interacting with that feature.
    Direct usage of high-level features from within :file:`_projectdesc` is intended to not use this Feature objects
    but to use the embedded properties of the universe object.
    """

    def __init__(self, name: str, hookhandlers: List[Any], featuremodule: Any, hooks: Dict[str, Hook]):
        """
        :param name: The optional name of the feature.
        :param hookhandlers: All hook handlers this feature registers globally (it may do that later on in a function
                             as well).
        :param featuremodule: The feature module.
        :param hooks: List of Hook instances.
        """
        self.name = name
        self._hookhandlers = list(hookhandlers)
        self.featuremodule = featuremodule
        self.hooks = hooks


_features = []

_featuresources = []


def loadfeature(name: str) -> Feature:
    """
    Load a feature by name from the predefined sources (as optional command-line argument). If if is already loaded,
    it returns the existing one.
    """
    global _hooks
    for f in _features:
        if f.name == name:
            return f
    frompath = None
    for srcfile, srcname in _featuresources:
        if srcname == name:
            frompath = srcfile
            break
    if frompath:
        anise.utils.logging.logdebug(f"Load feature from path '{frompath}' to name '{name}'.")
        _hooks.append([])
        try:
            loader = importlib.machinery.SourceFileLoader(name, frompath)
            # TODO allow also usual import and have other hook mechansism here
            spec = importlib.util.spec_from_loader(loader.name, loader)
            fmodule = importlib.util.module_from_spec(spec)
            loader.exec_module(fmodule)
            for hook in _hooks[-1]:
                if name not in hook.provides:
                    hook.provides.append(name)
                oname = getattr(hook.function, "__name__", None)
                if oname:
                    oname = name + "." + oname
                    if oname not in hook.provides:
                        hook.provides.append(oname)
            hooks = {}
            for ds in dir(fmodule):
                dos = getattr(fmodule, ds)
                if isinstance(dos, Hook):
                    hooks[ds] = dos
            _feature = Feature(name, _hooks[-1], fmodule, hooks)
        finally:
            _hooks.pop(len(_hooks)-1)
    else:
        anise.utils.logging.logwarning(f"Unable to load feature '{name}'.")
        _feature = Feature(name, [], None, {})
    _features.append(_feature)
    _features.sort(key=lambda x: x.name)
    return _feature


_hooks = []


def hookhandler(hookobj: Hook, *,
                provides: Optional[Union[str, List[str]]] = None, requires: Optional[Union[str, List[str]]] = None,
                prepares: Optional[Union[str, List[str]]] = None) -> Callable:
    """
    Used by a feature module to register a handler for a hook globally. The result should be used as Python function
    decorator.

    :param hookobj: The hook.
    :param provides: A list of symbol names this hook participates to provide (dependency provider).
    :param requires: A list of symbol names this hook requires to be executed before.
    :param prepares: A list of symbol names this hook must prepend.
    """
    def _realdecorator(func):
        if len(_hooks) > 0:
            _hooks[-1].append(HookHandler(hookobj, func, provides, requires, prepares))
        return func
    return _realdecorator


def getfeatures() -> List[Feature]:
    """
    Returns a list of all features.
    """
    return list(_features)


#: This hook gets executed by the Anise engine right before the :file:`_projectdesc` file is loaded.
HOOK_BEFORE_DEFINITION = Hook(doc="This hook is called before the project description file is loaded.")

#: This hook gets executed by the Anise engine after the :file:`_projectdesc` file is loaded and right before the
#: chosen task itself is called. At this time all the assignments and global statements in the :file:`_projectdesc`
#: are already applied.
HOOK_BEFORE_EXECUTION = Hook(doc="This hook is called after the project description file is loaded and custom "
                                 "initialization in HOOK_AFTER_PREPARATION took place.")


if TYPE_CHECKING:
    import anise.framework.projects as projects


"""TODO# anise 5

- features = usual python modules/packages
- features
  - documentation
  - base
  - versioncontrol
    - support git commit amend
    - minimalize automated commits
    - no more implicit invasive automated actions
  - dependencies
  - milieu
  - packages/distributables/packagestore
  - build
  - testing
    - no fixed testpackage:test 2 level hierarchy, but just a tree
  - filesystem/files/filetransfer
  - releasing
    - not have this single special task
  - files
    - more caching, easier composition
    - easier api; fix .dl()
  - homepage
- how to combine task implementation for (a) automated by parameters and (b) with user interaction
- fix 'ExecutionScope's?!
- general
  - add type annotations
- cleanup anise packages tree (utils, framework, ...) ?!
- richer (but simpler, cleaner) ways to steer and inject things
  - hooks, tasks, ...
  - low-level hook events (for infiltrating all hook triggerings in general)
  - oref alternatives?!
  - HOOK_BEFORE_DEFINITION in scripts inside the file: chicken-egg-problem -> one more hook?!
- richer task chooser; categories, ... ?!
- ci pipeline functionality?!
  - build time/success statistics
- GUI
  - richer user interaction (userfeedback)
  - projectdescriptioneditor
    - differently; not as separated features
  - task chooser
  - console
  - raw editor  
- _projectdesc
  - easier way to use small code pieces anywhere
  - allow multi-file _projectdesc: includes; more elaborated _meta dir structure?
  - shorter, more compact syntax
    - maybe json?! yaml?!
    - or use xml namespaces for some stuff?!
  - have .xml (or whatever) suffix
  - no val/let difference ?!
  - format strings; like f"foo{bar}"
  - shorter notation e.g. for base.set_basecolor ?!






































"""