# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Implementation for the Anise universe object and some services around it.
"""

from typing import *
import importlib
# noinspection PyPep8Naming
import xml.etree.cElementTree as ET

import anise.data.PyDepsEngine.depsengine
import anise.framework.exceptions
import anise.framework.features
import anise.utils


_currentuniverse = []


def setcurrentuniverse(u: 'Universe') -> None:
    """
    Sets the current universe object. If another one is currently set, it becomes active again when :samp:`u` is unset.
    |infrastructure|
    """
    _currentuniverse.append(u)


def unsetcurrentuniverse() -> None:
    """
    Unsets the current universe object.
    |infrastructure|
    """
    _currentuniverse.pop()


def getvalue(key: str, defaultvalue: Optional[Any] = None) -> Optional[Any]:
    """
    A convenience method returning the value of a certain property for the current universe object.

    :param key: The key name of the property to get.
    :param defaultvalue: This value is returned if the specified property does not exist.
    """
    return currentuniverse().getvalue(key, defaultvalue)


def setvalue(key: str, value: Optional[Any]) -> bool:
    """
    A convenience method setting the value of a certain property for the current universe object.

    :param key: The key name of the property to set.
    :param value: The new value for the specified property.
    :return: If successful.
    """
    return currentuniverse().setvalue(key, value)


def currentuniverse() -> Optional['Universe']:
    """
    Returns the current universe object.
    """
    return _currentuniverse[-1] if len(_currentuniverse) > 0 else None


class _UniverseProxy:
    """
    Internally used for the universe object for allowing just to write :samp:`universe` instead of :samp:`universe()`.
    """

    def __getattr__(self, item):
        return getattr(currentuniverse(), item)

    def __setattr__(self, key, value):
        setattr(currentuniverse(), key, value)

    def __call__(self, *args, **kwargs):  # compatibility with existing code
        return currentuniverse()


#: The universe object. See :py:class:`Universe`.
universe = _UniverseProxy()


class Universe:
    """
    Implementation for universe objects.
    """

    def __init__(self, d: Dict[str, Optional[Any]], executionscope: Optional['report.ExecutionScope'] = None):
        self.executionscope = executionscope
        self.append(d)
        self.hooks = {}
        self._registeredtasks = {}

    class RegisteredTask:
        """
        A task that is registered to the universe object. You typically do not need to work with this structure
        directly. Use :py:meth:`Universe.addtask` to add new tasks.
        """

        def __init__(self, parentname: str, taskname: str, label: Optional[str], important: bool, invisible: bool,
                     sortindex: int, function: Callable):
            self.parentname = parentname
            self.taskname = taskname
            self.name = ((parentname + ".") if parentname else "") + taskname
            self.label = label
            self.important = important
            self.invisible = invisible
            self.sortindex = sortindex
            self.function = function

    def addtask(self, function: Callable, name: str = "", *, label: Optional[str] = None, important: bool = True,
                invisible: bool = False, sortindex: int = 0) -> None:
        """
        Registers a task and makes it available for execution.

        :param function: The function that implements the task.
        :param name: The task name.
        :param label: The task label text.
        :param important: If this is an important/interesting task (presented differently in choosers).
        :param invisible: If this task should be hidden in choosers.
        :param sortindex: A number that controls the display order of tasks in choosers.
        """
        if not name or name.endswith("."):
            if not getattr(function, "__name__", None):
                raise anise.framework.exceptions.BadInputError(
                    "'function' must have a '__name__' attribute or 'name' must be a complete function name.")
            name += function.__name__
        ildot = name.rfind(".")
        if ildot > -1:
            parentname = name[:ildot]
            taskname = name[ildot+1:]
            parent = eval(parentname, self.__dict__)
        else:
            parentname = None
            taskname = name
            parent = self
        setattr(parent, taskname, function)
        self._registeredtasks[name] = self.RegisteredTask(parentname, taskname, label, important, invisible, sortindex,
                                                          function)

    def registeredtasks(self) -> List['Universe.RegisteredTask']:
        """
        Returns a list of all registered tasks.
        """
        return list(self._registeredtasks.values())

    def append(self, d: Dict[str, Optional[Any]]) -> None:
        """
        Stores a dictionary of values as properties of this universe object.
        """
        class DataStructure:
            pass
        for key in d:
            o = self
            parts = key.split(".")
            for kkey in parts[:-1]:
                ino = getattr(o, kkey, None)
                if ino is None:
                    ino = DataStructure()
                    setattr(o, kkey, ino)
                o = ino
            setattr(o, parts[-1], d[key])

    def addhook(self, hookobj: anise.framework.features.Hook, fct: Callable,
                params: Optional[Dict[str, Optional[Any]]] = None, provides: Optional[Union[str, List[str]]] = None,
                requires: Optional[Union[str, List[str]]] = None,
                prepares: Optional[Union[str, List[str]]] = None) -> anise.framework.features.HookHandler:
        """
        Adds a handler for a hook to the universe object.
        """
        if params is None:
            params = {}
        if callable(fct):
            nfct = (lambda *a, **b: fct(*a, **b, **params))
        else:
            nfct = fct
        hook = anise.framework.features.HookHandler(hookobj, nfct, provides, requires, prepares)
        hooklst = self.hooks.get(hookobj, None)
        if not hooklst:
            hooklst = self.hooks[hookobj] = anise.data.PyDepsEngine.depsengine.Engine()
        hooklst.add_object(hook, own=provides, afteroptional=requires, beforeoptional=prepares)
        return hook

    def removehook(self, hookhandler: anise.framework.features.HookHandler) -> None:
        """
        Removes a hook that was added before.

        :param hookhandler: The HookHandler you got in :py:meth:`addhook`.
        """
        hooklst = self.hooks.get(hookhandler.hookobj, None)
        if hooklst:
            hooklst.remove_object(hookhandler)

    def gethooks(self, hookobj: anise.framework.features.Hook) -> List[Callable]:
        """
        Gets the handlers for a hook in the correct order.

        :param hookobj: The hook you want to get handlers for.
        """
        hooks = self.hooks.get(hookobj, None)
        return [x.function for x in (hooks.get_objects() if hooks else [])]

    def getvalue(self, key: str, defaultvalue: Optional[Any] = None) -> Optional[Any]:
        """
        Returns the value of a certain property for the current universe object.
        
        If this property does not exist, it returns the :samp:`defaultval`.
        Beyond that, :samp:`getvalue("foo.bar")` is similar to :samp:`universe.foo.bar`.

        :param key: The key name of the property to get.
        :param defaultvalue: This value is returned if the specified property does not exist.
        """
        x = self
        keysegs = key.split(".")
        notfound = object()
        if (len(keysegs) > 0) and (keysegs[0] in ["universe", "universe()"]):
            keysegs.pop(0)
        for k in keysegs:
            if k:
                x = getattr(x, k, notfound)
                if x == notfound:
                    return defaultvalue
        return x

    def setvalue(self, key: str, value: Optional[Any]) -> bool:
        """
        Sets the value of a certain property for the current universe object.
        
        If a parent property does not exist, it does nothing.
        :samp:`setvalue("foo.bar", "baz")` is similar to :samp:`universe.foo.bar = "baz"`.

        :param key: The key name of the property to set.
        :param value: The new value for the specified property.
        :return: If successful.
        """
        keysegs = [k for k in key.split(".") if k]
        if len(keysegs) > 0:
            x = self.getvalue(".".join(keysegs[:-1]), None)
            if x:
                setattr(x, keysegs[-1], value)
                return True


class IntermediateStructure:
    """
    A data structure that holds an in-memory representation of the project description.
    It is not the in-memory representation of the live universe object state.
    It is like the file structure of the project description file, but not exactly the same.
    There is a parser and a generator for translating between a project description file and an IntermediateStructure.
    """

    class ParseProjectDescriptionError(anise.framework.exceptions.BadFormatError):
        """
        Parsing error like bad input xml.
        """
        pass

    class ParseProjectDescriptionInternalError(anise.framework.exceptions.InternalError):
        """
        Parsing error that seems like a bug.
        """
        pass

    class Node:
        """
        A node in the intermediate structure.
        """

        def __init__(self, k: str, **args):
            """
            :param k: The node key.
            """
            self._parent = None
            self._args = {}
            self._key = k
            for arg in args:
                self.setarg(arg, args[arg])
            self.islet = False

        @property
        def key(self) -> str:
            """
            The key name.
            """
            return self._key

        @property
        def parent(self) -> Union['IntermediateStructure.Node', 'IntermediateStructure']:
            """
            The parent node.

            The root node returns its :py:class:`IntermediateStructure`.
            """
            return self._parent

        @property
        def isvalue(self) -> bool:
            """
            If this is a value node. See also :py:class:`IntermediateStructure.Value`.
            """
            return False

        def args(self) -> List[str]:
            """
            Returns the list of stored argument keys.
            """
            return list(self._args.keys())

        def getarg(self, key: str) -> 'IntermediateStructure.Value':
            """
            Returns an argument.

            :param key: The argument key.
            """
            return self._args.get(key, None)

        def setarg(self, key: str, node: 'IntermediateStructure.Value') -> None:
            """
            Sets an argument.

            :param key: The argument key.
            :param node: The argument value.
            """
            if not isinstance(node, IntermediateStructure.Value):
                raise anise.framework.exceptions.BadInputError("Not an IntermediateStructure.Value: "+repr(node))
            node._parent = self
            node._key = key
            oldnode = self.getarg(key)
            if oldnode:
                oldnode.remove()
            self._args[key] = node

        def removearg(self, key: str) -> 'IntermediateStructure.Value':
            """
            Removes an argument.

            :param key: The argument key.
            """
            return self._args.pop(key)

        def clearargs(self) -> None:
            """
            Clears the list of arguments.
            """
            self._args.clear()

        def remove(self) -> None:
            """
            Removes this node.
            """
            if isinstance(self.parent, IntermediateStructure):
                self.parent.removeentry(self)
            else:
                self.parent.removearg(self.key)

    class Value(Node):
        """
        A Value entry in the project description structure. This corresponds to a :samp:`val` node in the xml format.
        It represents an arbitrary value (primitive or composed type).
        """

        def __init__(self, t: str, *, k: Optional[str] = None, v: Optional[str] = None, oref: Optional[str] = None,
                     **args):
            """
            :param t The value type.
            :param k The key (i.e. property name in parent object).
            :param v (optional) The value string for a primitive value.
            :param oref (optional) The oref string for a objectref value.
            """
            super().__init__(k, **args)
            self.__vtype = t
            self.__value = v
            self.__oref = oref

        @property
        def isvalue(self):
            return True

        @property
        def value(self) -> str:
            """
            The value (for primitive values).
            """
            return self.__value

        @value.setter
        def value(self, v: Optional[Any]):
            vv = str(v)  # value checking
            eval(self.__vtype)(vv)
            self.__value = vv if (v is not None) else None

        @property
        def vtype(self) -> str:
            """
            The value type name.
            """
            return self.__vtype

        @vtype.setter
        def vtype(self, t):
            self.__vtype = t

        @property
        def oref(self) -> str:
            """
            The objectref (for oref values).
            """
            return self.__oref

        @oref.setter
        def oref(self, v):
            self.__oref = v

        def appendarg(self, node: 'IntermediateStructure.Value') -> None:
            """
            Appends a new value node to a list node.
            """
            if self.vtype != "list":
                raise anise.framework.exceptions.BadInputError("appendarg requires a list, but got: "+repr(self))
            self.setarg("i9999999", node)
            self._realign_listvalues()

        def _realign_listvalues(self) -> None:
            """
            Makes internal list realignment tasks.

            This is required to call after the list structure changes.
            """
            tmp = {}
            for i, x in enumerate(sorted(self.args())):
                tmp["i" + f"00000{i}"[:6] + "0"] = self.getarg(x)
            self.clearargs()
            for x in tmp:
                self.setarg(x, tmp[x])

        def getprimitiverepr(self) -> Optional[str]:
            """
            Returns a string representation of the value, if it is a primitive one (:samp:`None` otherwise).
            """
            if self.vtype == 'objectref':
                return self.oref
            elif self.value is None:
                return None
            elif self.vtype == 'bool':
                return str(bool(IntermediateStructure.Value.istrue(self.value)))
            elif self.vtype in ['str', 'int', 'float']:
                return repr(self.value)

        @staticmethod
        def istrue(s: str) -> bool:
            """
            Returns :samp:`True` iff the given string can represent a boolean :samp:`True` value.
            """
            return s.lower() in ["1", "y", "true"]

        @staticmethod
        def fromstring(val: Any, key: Optional[str] = None) -> 'IntermediateStructure.Value':
            """
            Convenience method that creates a string Value.
            """
            return IntermediateStructure.Value("str", k=key, v=str(val))

        @staticmethod
        def frombool(val: Any, key: Optional[str] = None) -> 'IntermediateStructure.Value':
            """
            Convenience method that creates a boolean Value.
            """
            return IntermediateStructure.Value("bool", k=key, v=str(bool(val)))

        def __str__(self):
            if self.vtype in ["str", "int", "bool", "float"]:
                return str(self.value)
            elif self.vtype == "objectref":
                return str(self.oref)
            elif self.vtype == "list":
                return str([str(self.getarg(x)) for x in sorted(self.args())])
            elif self.vtype == "dict":
                return str({x: str(self.getarg(x)) for x in self.args()})
            else:
                return "['{0}' ({1}) {2}]".format(self.key, self.vtype, "; ".join(
                                                    [x + "=" + str(self.getarg(x)) for x in sorted(self.args())]))

    class Let(Node):
        """
        A Let entry in the project description structure. This corresponds to a :samp:`let` node in the xml format.
        It represents the execution of a parameterized part of anise feature program logic.
        """

        def __init__(self, k: str, **args):
            """
            :param k The full qualified method name.
            """
            super().__init__(k, **args)
            self.islet = True

        def __str__(self):
            return "[LET '{0}' {1}]".format(self.key, "; ".join([x+"="+str(self._args[x])
                                                                   for x in sorted(self._args)]))

    def __init__(self, entries: Optional[List['IntermediateStructure.Node']] = None) -> None:
        """
        :param entries: The new list of direct subitems.
        """
        self._entries = []
        self.sourcefilepath = None
        self.sourceformat = None
        if entries:
            for ent in entries:
                self.addentry(ent)

    def entries(self) -> List['IntermediateStructure.Node']:
        """
        Returns the list of direct subitems.
        """
        return list(self._entries)

    def addentry(self, node: 'IntermediateStructure.Node') -> None:
        """
        Adds a new entry to the list of direct subentries.

        :param node: The new node.
        """
        self._entries.append(node)
        node._parent = self

    def removeentryat(self, i: int) -> 'IntermediateStructure.Node':
        """
        Removes an entry from the list of direct subentries.

        :param i: The entry position to remove.
        :return: The deleted node.
        """
        return self._entries.pop(i)

    def removeentry(self, node: 'IntermediateStructure.Node') -> None:
        """
        Removes an entry from the list of direct subentries.

        :param node: The child node to remove.
        """
        self._entries.remove(node)

    def setentries(self, entries: List['IntermediateStructure.Node']) -> None:
        """
        Sets the list of direct subentries.

        :param entries: The new list of nodes.
        """
        self._entries.clear()
        for ent in entries:
            self.addentry(ent)

    def __str__(self):
        return "[IS\n{0}\n]".format("\n  ".join([str(x) for x in self.entries()]))

    @staticmethod
    def fromxml(xmlstring: str, sourcefilepath: Optional[str] = None) -> 'IntermediateStructure':
        """
        Generates an IntermediateStructure for an xml document.

        :param xmlstring: XML content.
        :param sourcefilepath: Path to the source file, e.g. used for :py:meth:`save`.
        """
        entries = []
        xdoc = ET.ElementTree(ET.fromstring(xmlstring))
        xroot = xdoc.getroot()
        for xentry in xroot:
            if xentry.tag == "val":
                entries.append(IntermediateStructure._fromxml_val(xentry))
            elif xentry.tag == "let":
                entries.append(IntermediateStructure._fromxml_let(xentry))
            else:
                raise IntermediateStructure.ParseProjectDescriptionError("invalid tag: "+xentry.tag)
        result = IntermediateStructure(entries=entries)
        result.sourcefilepath = sourcefilepath
        result.sourceformat = "xml"
        return result

    @staticmethod
    def fromfile(filename: str) -> 'IntermediateStructure':
        """
        Generates an IntermediateStructure for an xml file.

        :param filename: Path to the project description file.
        """
        # we currently assume xml here since it is the only format
        result = IntermediateStructure.fromxml(anise.utils.basic.readfromfile(filename), filename)
        return result

    def tofile(self, filename: str) -> None:
        """
        Saves the current state of this intermediate structure to a file.

        :param filename: The destination path.
        """
        if self.sourceformat == "xml":
            content = self.toxml()
        else:
            raise anise.framework.exceptions.BadInputError(
                "Unknown project description file format: " + repr(self.sourceformat))
        anise.utils.basic.writetofile(filename, content)

    def save(self) -> None:
        """
        Saves the current state of this intermediate structure back to the disk.

        This typically works only for instances that were loaded from disk.
        """
        self.tofile(self.sourcefilepath)

    def toxml(self) -> str:
        """
        Generates an xml representation for this intermediate structure.
        """
        xroot = ET.Element("aniseprojectdesc")
        for entry in self.entries():
            if isinstance(entry, IntermediateStructure.Value):
                tag = IntermediateStructure._toxml_val(entry, None)
            elif isinstance(entry, IntermediateStructure.Let):
                tag = IntermediateStructure._toxml_let(entry)
            else:
                raise IntermediateStructure.ParseProjectDescriptionInternalError("Unknown thing: "+str(entry))
            xroot.append(tag)
        import xml.dom.minidom  # argh
        return xml.dom.minidom.parseString(ET.tostring(xroot)).toprettyxml(indent="    ")

    def find(self, key: str, lets: bool = False, vals: bool = False) -> List['IntermediateStructure.Node']:
        """
        Find all :samp:`let`s and/or :samp:`val`s with a certain key.
        """
        r = [x for x in self.entries() if x.key == key]
        if not lets:
            r = [x for x in r if not isinstance(x, IntermediateStructure.Let)]
        if not vals:
            r = [x for x in r if not isinstance(x, IntermediateStructure.Value)]
        return r

    def inject_to_universe(self, u: 'Universe', g: Dict[str, Optional[Any]], l: Dict[str, Optional[Any]]) -> None:
        """
        Injects the (always static) information from this project description into a living universe object
        (typically for executing a task on top of it).

        :param u: The universe object.
        :param g: The globals dict.
        :param l: The locals dict.
        """
        for entry in self.entries():
            if isinstance(entry, IntermediateStructure.Value):
                x = IntermediateStructure._inject_to_universe_value(entry, g, l)
                u.append({entry.key: x})
            elif isinstance(entry, IntermediateStructure.Let):
                IntermediateStructure._inject_to_universe_let(entry, g, l)
            else:
                raise IntermediateStructure.ParseProjectDescriptionInternalError(f"Unknown thing: {entry}")

    @staticmethod
    def _inject_to_universe_let(entry: 'IntermediateStructure.Let',
                                g: Dict[str, Optional[Any]], l: Dict[str, Optional[Any]]) -> None:
        """
        Executes the let action from a let node.
        """
        fct = IntermediateStructure._inject_to_universe_resolve(entry.key, g, l)
        nparams = {}
        for y in entry.args():
            nparams[y] = IntermediateStructure._inject_to_universe_value(entry.getarg(y), g, l)
        fct(**nparams)

    @staticmethod
    def resolveobjectref(oref: str) -> Optional[Any]:
        """
        Returns the resolved object for an objectref string.

        :param oref: The objectref string (e.g. :samp:`'homepage.makehomepage'` or :samp:`'some.foo.Class().meth()+1'`).
        """
        return IntermediateStructure._inject_to_universe_resolve(
            oref, dict(importlib.import_module("anise.framework.imports").__dict__), universe().__dict__)

    @staticmethod
    def _inject_to_universe_resolve(oref: str,
                                    g: Dict[str, Optional[Any]], l: Dict[str, Optional[Any]]) -> Optional[Any]:
        """
        Returns the resolved object for an objectref string.
        """
        from anise.framework import imports
        if hasattr(imports, oref):
            return getattr(imports, oref)
        return eval(oref, g, l)

    @staticmethod
    def _inject_to_universe_value(entry: 'IntermediateStructure.Value',
                                  g: Dict[str, Optional[Any]], l: Dict[str, Optional[Any]]) -> Optional[Any]:
        """
        Returns a plain Python value from a value node.
        """
        if entry.vtype == "str":
            result = entry.value
        elif entry.vtype == "int":
            result = int(entry.value) if entry.value else 0
        elif entry.vtype == "bool":
            result = IntermediateStructure.Value.istrue(entry.value)
        elif entry.vtype == "float":
            result = float(entry.value) if entry.value else 0.0
        elif entry.vtype == "objectref":
            result = IntermediateStructure._inject_to_universe_resolve(entry.oref, g, l)
        elif entry.vtype == "list":
            result = []
            for k in sorted(entry.args()):
                result.append(IntermediateStructure._inject_to_universe_value(entry.getarg(k), g, l))
        elif entry.vtype == "dict":
            result = {}
            for k in entry.args():
                result[k] = IntermediateStructure._inject_to_universe_value(entry.getarg(k), g, l)
        else:
            ptype = IntermediateStructure._inject_to_universe_resolve(entry.vtype, g, l)
            args = {}
            for k in entry.args():
                args[k] = IntermediateStructure._inject_to_universe_value(entry.getarg(k), g, l)
            result = ptype(**args)
        return result

    @staticmethod
    def _fromxml_val(xentry: ET.Element) -> 'IntermediateStructure.Value':
        """
        Creates a value node from an xml node.
        """
        oargs = dict(xentry.attrib)
        if "t" in oargs:
            vtype = oargs["t"]
            del oargs["t"]
        elif IntermediateStructure._fromxml_val_isstring(xentry):
            vtype = "str"
        elif IntermediateStructure._fromxml_val_isdict(xentry):
            vtype = "dict"
        elif IntermediateStructure._fromxml_val_islist(xentry):
            vtype = "list"
        elif "oref" in oargs:
            vtype = "objectref"
        else:
            raise IntermediateStructure.ParseProjectDescriptionError("no 't' in '{0} {1}'".format(
                                    xentry.tag, xentry.attrib))
        result = IntermediateStructure.Value(vtype, k=xentry.attrib.get("k", None))
        if "k" in oargs:
            del oargs["k"]
        if (vtype == "str") and ("v" not in xentry.attrib):
            oargs["v"] = xentry.text
        if vtype in ["str", "int", "bool", "float"]:
            args = oargs
            result.value = args.pop("v")
        elif vtype == "objectref":
            args = oargs
            result.oref = args.pop("oref")
        else:
            args = {}
            for y in oargs:
                args[y] = IntermediateStructure.Value.fromstring(oargs[y], key=y)
        for ichild, xchild in enumerate(xentry):
            if vtype == "list":
                key = "i" + ("00000" + str(ichild))[:6] + "0"
            else:
                if "k" not in xchild.attrib:
                    raise IntermediateStructure.ParseProjectDescriptionError("no 'k' in '{0} {1}'".format(
                                            xchild.tag, xchild.attrib))
                key = xchild.attrib["k"]
            args[key] = IntermediateStructure._fromxml_val(xchild)
        for y in args:
            result.setarg(y, args[y])
        return result

    @staticmethod
    def _fromxml_val_isstring(xentry: ET.Element) -> bool:
        """
        Checks if a particular xml node represents a string.
        """
        if len([y for y in xentry.attrib.keys() if y not in ["k", "t", "v"]]) > 0:
            return False
        return len(xentry) == 0

    @staticmethod
    def _fromxml_val_isdict(xentry: ET.Element, withoutkeys: bool = False) -> bool:
        """
        Checks if a particular xml node represents a value dict.
        """
        if len([y for y in xentry.attrib.keys() if y not in ["k", "t"]]) > 0:
            return False
        cl = list(xentry)
        if len(cl) == 0:
            return False
        return len([c for c in cl if ("k" in c.attrib) == withoutkeys]) == 0

    @staticmethod
    def _fromxml_val_islist(xentry: ET.Element) -> bool:
        """
        Checks if a particular xml node represents a value list.
        """
        return IntermediateStructure._fromxml_val_isdict(xentry, withoutkeys=True)

    @staticmethod
    def _fromxml_let(xentry: ET.Element) -> 'IntermediateStructure.Let':
        """
        Creates a let node from an xml node.
        """
        rparams = dict(xentry.attrib)
        del rparams["k"]
        params = {}
        for y in rparams:
            params[y] = IntermediateStructure.Value.fromstring(rparams[y], y)
        for xchild in xentry:
            if "k" not in xchild.attrib:
                raise IntermediateStructure.ParseProjectDescriptionError("no 'k' in '{0} {1}'".format(
                                        xchild.tag, xchild.attrib))
            params[xchild.attrib["k"]] = IntermediateStructure._fromxml_val(xchild)
        return IntermediateStructure.Let(xentry.attrib["k"], **params)

    @staticmethod
    def _toxml_val(entry: 'IntermediateStructure.Value', xcurr) -> ET.Element:
        """
        Converts a value node to an xml node.
        """
        xresult = ET.Element("val")
        if not(isinstance(entry.parent, IntermediateStructure.Value) and entry.parent.vtype == "list"):
            xresult.attrib["k"] = str(entry.key)
        if entry.vtype == "str":
            txt = str(entry.value)
            if (len(txt) < 75) and ("\n" not in txt):
                if (xcurr is not None) and ("k" in xresult.attrib) and (len(txt)+len(xresult.attrib["k"])+4) <= 80:
                    xcurr.attrib[xresult.attrib["k"]] = txt
                    xresult = None
                else:
                    xresult.attrib["v"] = txt
            else:
                xresult.text = txt
        elif entry.vtype in ["int", "bool", "float"]:
            xresult.attrib["v"] = str(entry.value)
            xresult.attrib["t"] = str(entry.vtype)
        elif entry.vtype == "objectref":
            xresult.attrib["oref"] = str(entry.oref)
        else:
            if (entry.vtype in ["dict", "list"]) and len(entry.args()) > 0:
                pass
            else:
                xresult.attrib["t"] = str(entry.vtype)
            for y in sorted(entry.args()):
                xnn = IntermediateStructure._toxml_val(entry.getarg(y), xresult if (entry.vtype
                                                                                    not in ["list", "dict"]) else None)
                if xnn is not None:
                    xresult.append(xnn)
        return xresult

    @staticmethod
    def _toxml_let(entry: 'IntermediateStructure.Let') -> ET.Element:
        """
        Converts a let node to an xml node.
        """
        xresult = ET.Element("let")
        xresult.attrib["k"] = str(entry.key)
        for param in sorted(entry.args()):
            xparam = IntermediateStructure._toxml_val(entry.getarg(param), xresult)
            if xparam is not None:
                xresult.append(xparam)
        return xresult


if TYPE_CHECKING:
    import anise.framework.report as report
