# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Basic helpers for filesystem and other operating system related stuff.
"""

from typing import *
import os
import subprocess
import sys
import threading
import time

import anise.framework.exceptions
import anise.utils.logging


_throw = object()


def writetofile(filename: str, content: AnyStr) -> None:
    """
    Writes some content to a file.

    :param filename: The file path.
    :param content: The content to write.
    """
    with open(filename, "wb" if isinstance(content, bytes) else "w") as f:
        f.write(content)


def readfromfile(filename: str, onestring: bool = True, notexist_default: Any = _throw, binary: bool = False) -> AnyStr:
    """
    Reads the content from a file.

    :param filename: The file path.
    :param onestring: If the content is to be read in one single string (instead of a list of lines).
    :param notexist_default: The default content when the file does not exist.
    :param binary: If this is binary content instead of text.
    """
    if (notexist_default != _throw) and not os.path.exists(filename):
        return notexist_default
    with open(filename, "rb" if binary else "r") as f:
        if binary:
            r = f.read()
        else:
            r = f.readlines()
            if onestring:
                r = "".join(r)
    return r


def mkdirs(dname: str) -> None:
    """
    Creates a directory path recursively if it does not yet exist.

    :param dname: The directory path.
    """
    if not os.path.isdir(dname):
        os.makedirs(dname)


def maketarball(sourcepath, tarpath: str) -> None:
    """
    Makes a tar archive from a given directory structure.

    :param sourcepath: The root source directory of the new tarball (becomes a directory in the archive as typical for
                       tarballs).
    :param tarpath: The destination path for the tar archive.
    """
    sourcepath = os.path.abspath(sourcepath)
    with ChDir(f"{sourcepath}/.."):
        call(["tar", "cfz", tarpath, os.path.basename(sourcepath)], raise_on_errors="Errors creating a tarball")


def call(*cmdline: Union[str, List[str]], shell: bool = False, decode: bool = True,
         raise_on_errors: Optional[str] = None,
         split_output_channels: bool = False) -> Tuple[int, Union[AnyStr, Tuple[AnyStr, AnyStr]]]:
    """
    Executes a command line.

    :param cmdline: The command line as list of arguments.
    :param shell: If the command shall be interpreted by a shell (only the first part of cmdline will apply!).
    :param decode: If the result is to be decoded to a string.
    :param raise_on_errors: If errors shall raise Exceptions (also specifies the exception text in this case).
    :param split_output_channels: If to separately record output and error stream.
    :return: Tuple of returncode, program output.
    """
    ret = 0
    try:
        if len(cmdline) == 1:
            cmdline = cmdline[0]
        r = subprocess.check_output(cmdline, stderr=subprocess.STDOUT if (not split_output_channels) else None,
                                    shell=shell)
        er = b""
    except subprocess.CalledProcessError as err:
        ret = err.returncode
        r = err.output if (not split_output_channels) else err.stdout
        er = err.stderr or b""
    except OSError as err:
        ret = -1
        r = b""
        er = str(err).encode()
    if raise_on_errors and (ret != 0):
        raise anise.framework.exceptions.ProcessExecutionFailedError(
            str(raise_on_errors) + ": " + str(cmdline), r.decode(errors='replace').strip())
    if decode:
        r = r.decode().strip()
        er = er.decode().strip()
    return ret, ((r, er) if split_output_channels else r)


def callanise(*args: str) -> None:
    """
    Calls the Anise process (in a platform independent way).

    :param args: The command line arguments to pass.
    """
    cmdline = []
    if sys.platform == "win32":
        cmdline.append(sys.executable)
    cmdline.append(sys.argv[0])
    cmdline += args
    subprocess.call(cmdline)


def verify_tool_installed(toolname: str) -> None:
    """
    Checks if a tool is installed and raises an exception if not.

    :param toolname: The tool name as to execute on command line.
    """
    try:
        subprocess.check_output([toolname, '--version'], stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        pass
    except IOError:
        raise anise.framework.exceptions.RequirementsMissingError(f"Required tool '{toolname}' not found.")


class Mount:
    """
    Mounts a volume. Can either be used automatically with the :samp:`with` keyword or manually.
    """

    def __init__(self, src: str, dst: str, options: Optional[List[str]] = None, mounter: Optional[List[str]] = None,
                 unmounter: Optional[List[str]] = None, needsroot: bool = True):
        """
        :param src: Mount source.
        :param dst: Mount destination.
        :param options: Optional additional mount options.
        :param mounter: Optional string list overriding the default mount call.
        :param unmounter: Optional string list overriding the default unmount call.
        :param needsroot: If this mount action will need root permissions.
        """
        self.su = ["su", "-c"] if needsroot else []
        self.mounted = False
        self.src = src
        self.dst = dst
        self.options = options or []
        self.mounter = mounter or ["mount"]
        self.unmounter = unmounter or ["umount"]
        self.needsroot = needsroot

    def mount(self) -> None:
        """
        Manually mount the volume.
        """
        self.su = ["su", "-c"] if self.needsroot else []
        call(*(self.su + self.mounter + [self.src, self.dst] + self.options),
             raise_on_errors="unable to mount '{src}' to '{dst}'.".format(src=self.src, dst=self.dst))
        self.mounted = True

    def __enter__(self):
        self.mount()

    def unmount(self) -> None:
        """
        Manually unmount the volume.
        """
        if self.mounted:
            subprocess.call(["sync"])
            subprocess.call(["sync"])
            subprocess.call(["sync"])
            time.sleep(1)
            subprocess.call(self.su + self.unmounter + [self.dst])
            self.mounted = False

    def __exit__(self, etype, evalue, etraceback):
        self.unmount()


class ChDir:
    """
    Temporarily changes the current working directory in a scope of the :samp:`with` keyword.
    """

    def __init__(self, path: str):
        """
        :param path: The path to the new working directory.
        """
        self.newpath = path

    def __enter__(self):
        self.oldpath = os.getcwd()
        os.chdir(self.newpath)
        return self

    def __exit__(self, etype, evalue, etraceback):
        os.chdir(self.oldpath)


def execwithretry(fct: Callable, *args: Optional[Any], maxtries: int = 10, delaysecs: float = 0.2,
                  **kwargs: Optional[Any]) -> Optional[Any]:
    """
    Executes a function with a retry mechanism (executing it again when an exception was raised).

    :param fct: The function to execute.
    :param args: Arguments for the function call.
    :param maxtries: Maximum try count.
    :param delaysecs: Delay (in seconds) betweens tries.
    :param kwargs: Keyword arguments for the function call.
    """
    while True:
        maxtries -= 1
        try:
            res = fct(*args, **kwargs)
            break
        except Exception:
            if maxtries <= 0:
                raise
            time.sleep(delaysecs)
    return res


def openurl(url: str) -> None:
    """
    Opens a url (or file path) in the operating system's default way.

    :param url: A url or file path.
    """
    if sys.platform == "win32":
        # noinspection PyUnresolvedReferences
        threading.Thread(target=lambda: os.startfile(url)).start()
    else:
        r, o = call("xdg-open --version", shell=True)
        if r == 0:
            threading.Thread(target=lambda: call("nohup", "xdg-open", url), daemon=True).start()
        else:
            anise.utils.logging.logwarning("Anise is not able to automatically open files on your system.\nPlease open "
                                           "the following file in a suitable program:\n  {}".format(url),
                                           printcaller=False)


def openfile(filepath: str) -> None:
    """
    Opens a file path in the operating system's default way.

    :param filepath: A file path.
    """
    openurl(filepath)


def getdriveletters() -> List[str]:
    """
    Returns a list of drive letters that exist on this system (only win32).
    """
    return [l for l in "abcdefghijklmnopqrstuvwxyz" if os.path.exists(l + ":")]
